# -*- encoding: utf-8 -*-
# Software Management Providers
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Just a common functionality related to LMI_SoftwareInstallationService
provider.
"""

import pywbem

from lmi.providers import cmpi_logging
from lmi.providers import ComputerSystem
from lmi.providers import is_this_system
from lmi.software import util
from lmi.software.core import Identity
from lmi.software.core import Job
from lmi.software.core import SystemCollection
from lmi.software.yumdb import errors
from lmi.software.yumdb import YumDB

JOB_METHOD_SRC_PARAM_NAMES = ["URI", "Source", "Image", "Source"]

LOG = cmpi_logging.get_logger(__name__)

class InstallationError(Exception):
    """This exception shall be raised upon any error within
    install_or_remove_package() function.
    """
    def __init__(self, return_code, description):
        Exception.__init__(self, return_code, description)
    @property
    def return_code(self):
        """@return return code of CIM method"""
        return self.args[0]
    @property
    def description(self):
        """@return description of error"""
        return self.args[1]

class Values(object):
    class DetailedStatus(object):
        Not_Available = pywbem.Uint16(0)
        No_Additional_Information = pywbem.Uint16(1)
        Stressed = pywbem.Uint16(2)
        Predictive_Failure = pywbem.Uint16(3)
        Non_Recoverable_Error = pywbem.Uint16(4)
        Supporting_Entity_in_Error = pywbem.Uint16(5)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Not Available',
                1: 'No Additional Information',
                2: 'Stressed',
                3: 'Predictive Failure',
                4: 'Non-Recoverable Error',
                5: 'Supporting Entity in Error'
        }

    class RequestedState(object):
        Unknown = pywbem.Uint16(0)
        Enabled = pywbem.Uint16(2)
        Disabled = pywbem.Uint16(3)
        Shut_Down = pywbem.Uint16(4)
        No_Change = pywbem.Uint16(5)
        Offline = pywbem.Uint16(6)
        Test = pywbem.Uint16(7)
        Deferred = pywbem.Uint16(8)
        Quiesce = pywbem.Uint16(9)
        Reboot = pywbem.Uint16(10)
        Reset = pywbem.Uint16(11)
        Not_Applicable = pywbem.Uint16(12)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 32768..65535
        _reverse_map = {
                0: 'Unknown',
                2: 'Enabled',
                3: 'Disabled',
                4: 'Shut Down',
                5: 'No Change',
                6: 'Offline',
                7: 'Test',
                8: 'Deferred',
                9: 'Quiesce',
                10: 'Reboot',
                11: 'Reset',
                12: 'Not Applicable'
        }

    class HealthState(object):
        Unknown = pywbem.Uint16(0)
        OK = pywbem.Uint16(5)
        Degraded_Warning = pywbem.Uint16(10)
        Minor_failure = pywbem.Uint16(15)
        Major_failure = pywbem.Uint16(20)
        Critical_failure = pywbem.Uint16(25)
        Non_recoverable_error = pywbem.Uint16(30)
        # DMTF_Reserved = ..
        # Vendor_Specific = 32768..65535
        _reverse_map = {
                0: 'Unknown',
                5: 'OK',
                10: 'Degraded/Warning',
                15: 'Minor failure',
                20: 'Major failure',
                25: 'Critical failure',
                30: 'Non-recoverable error'
        }

    class InstallFromURI(object):
        Job_Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unspecified_Error = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Target_In_Use = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        Method_Parameters_Checked___Job_Started = pywbem.Uint32(4096)
        Unsupported_TargetType = pywbem.Uint32(4097)
        Unattended_silent_installation_not_supported = pywbem.Uint32(4098)
        Downgrade_reinstall_not_supported = pywbem.Uint32(4099)
        Not_enough_memory = pywbem.Uint32(4100)
        Not_enough_swap_space = pywbem.Uint32(4101)
        Unsupported_version_transition = pywbem.Uint32(4102)
        Not_enough_disk_space = pywbem.Uint32(4103)
        Software_and_target_operating_system_mismatch = pywbem.Uint32(4104)
        Missing_dependencies = pywbem.Uint32(4105)
        Not_applicable_to_target = pywbem.Uint32(4106)
        URI_not_accessible = pywbem.Uint32(4107)
        # Method_Reserved = 4108..32767
        # Vendor_Specific = 32768..65535
        class InstallOptions(object):
            Defer_target_system_reset = pywbem.Uint16(2)
            Force_installation = pywbem.Uint16(3)
            Install = pywbem.Uint16(4)
            Update = pywbem.Uint16(5)
            Repair = pywbem.Uint16(6)
            Reboot = pywbem.Uint16(7)
            Password = pywbem.Uint16(8)
            Uninstall = pywbem.Uint16(9)
            Log = pywbem.Uint16(10)
            SilentMode = pywbem.Uint16(11)
            AdministrativeMode = pywbem.Uint16(12)
            ScheduleInstallAt = pywbem.Uint16(13)
            # DMTF_Reserved = ..
            # Vendor_Specific = 32768..65535
            _reverse_map = {
                2  : "Defer Target/System Reset",
                3  : "Force Installation",
                4  : "Install",
                5  : "Update",
                6  : "Repair",
                7  : "Reboot",
                8  : "Password",
                9  : "Uninstall",
                10 : "Log",
                11 : "SilentMode",
                12 : "AdministrativeMode",
                13 : "ScheduleInstallAt",
            }

        InstallOptions.supported = {
            InstallOptions.Install,
            InstallOptions.Update,
            InstallOptions.Uninstall,
            InstallOptions.Force_installation,
            InstallOptions.Repair,
        }


    class CheckSoftwareIdentity(object):
        Job_Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unspecified_Error = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Target_In_Use = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        Method_Reserved = pywbem.Uint32(4096)
        Unsupported_TargetType = pywbem.Uint32(4097)
        Unattended_silent_installation_not_supported = pywbem.Uint32(4098)
        Downgrade_reinstall_not_supported = pywbem.Uint32(4099)
        Not_enough_memory = pywbem.Uint32(4100)
        Not_enough_swap_space = pywbem.Uint32(4101)
        Unsupported_version_transition = pywbem.Uint32(4102)
        Not_enough_disk_space = pywbem.Uint32(4103)
        Software_and_target_operating_system_mismatch = pywbem.Uint32(4104)
        Missing_dependencies = pywbem.Uint32(4105)
        Not_applicable_to_target = pywbem.Uint32(4106)
        No_supported_path_to_image = pywbem.Uint32(4107)
        Cannot_add_to_Collection = pywbem.Uint32(4108)
        Asynchronous_Job_already_in_progress = pywbem.Uint32(4109)
        # Method_Reserved = 4110..32767
        # Vendor_Specific = 32768..65535
        class InstallCharacteristics(object):
            Target_automatic_reset = pywbem.Uint16(2)
            System_automatic_reset = pywbem.Uint16(3)
            Separate_target_reset_Required = pywbem.Uint16(4)
            Separate_system_reset_Required = pywbem.Uint16(5)
            Manual_Reboot_Required = pywbem.Uint16(6)
            No_Reboot_Required = pywbem.Uint16(7)
            User_Intervention_recommended = pywbem.Uint16(8)
            MAY_be_added_to_specified_Collection = pywbem.Uint16(9)
            # DMTF_Reserved = ..
            # Vendor_Specific = 0x7FFF..0xFFFF

    class ChangeAffectedElementsAssignedSequence(object):
        Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Error_Occured = pywbem.Uint32(2)
        Busy = pywbem.Uint32(3)
        Invalid_Reference = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Access_Denied = pywbem.Uint32(6)
        # DMTF_Reserved = 7..32767
        # Vendor_Specified = 32768..65535

    class TransitioningToState(object):
        Unknown = pywbem.Uint16(0)
        Enabled = pywbem.Uint16(2)
        Disabled = pywbem.Uint16(3)
        Shut_Down = pywbem.Uint16(4)
        No_Change = pywbem.Uint16(5)
        Offline = pywbem.Uint16(6)
        Test = pywbem.Uint16(7)
        Defer = pywbem.Uint16(8)
        Quiesce = pywbem.Uint16(9)
        Reboot = pywbem.Uint16(10)
        Reset = pywbem.Uint16(11)
        Not_Applicable = pywbem.Uint16(12)
        # DMTF_Reserved = ..
        _reverse_map = {
                0: 'Unknown',
                2: 'Enabled',
                3: 'Disabled',
                4: 'Shut Down',
                5: 'No Change',
                6: 'Offline',
                7: 'Test',
                8: 'Defer',
                9: 'Quiesce',
                10: 'Reboot',
                11: 'Reset',
                12: 'Not Applicable'
        }

    class EnabledDefault(object):
        Enabled = pywbem.Uint16(2)
        Disabled = pywbem.Uint16(3)
        Not_Applicable = pywbem.Uint16(5)
        Enabled_but_Offline = pywbem.Uint16(6)
        No_Default = pywbem.Uint16(7)
        Quiesce = pywbem.Uint16(9)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 32768..65535
        _reverse_map = {
                2: 'Enabled',
                3: 'Disabled',
                5: 'Not Applicable',
                6: 'Enabled but Offline',
                7: 'No Default',
                9: 'Quiesce'
        }

    class EnabledState(object):
        Unknown = pywbem.Uint16(0)
        Other = pywbem.Uint16(1)
        Enabled = pywbem.Uint16(2)
        Disabled = pywbem.Uint16(3)
        Shutting_Down = pywbem.Uint16(4)
        Not_Applicable = pywbem.Uint16(5)
        Enabled_but_Offline = pywbem.Uint16(6)
        In_Test = pywbem.Uint16(7)
        Deferred = pywbem.Uint16(8)
        Quiesce = pywbem.Uint16(9)
        Starting = pywbem.Uint16(10)
        # DMTF_Reserved = 11..32767
        # Vendor_Reserved = 32768..65535
        _reverse_map = {
                0: 'Unknown',
                1: 'Other',
                2: 'Enabled',
                3: 'Disabled',
                4: 'Shutting Down',
                5: 'Not Applicable',
                6: 'Enabled but Offline',
                7: 'In Test',
                8: 'Deferred',
                9: 'Quiesce',
                10: 'Starting'
        }

    class VerifyInstalledIdentity(object):
        Job_Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unspecified_Error = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Target_In_Use = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        Method_Parameters_Checked___Job_Started = pywbem.Uint32(4096)
        Unsupported_TargetType = pywbem.Uint32(4097)
        Method_Reserved = pywbem.Uint32(4098)
        # Software_Identity_Not_Installed = 4099..32767
        Vendor_Specific = pywbem.Uint32(32768)

    class AvailableRequestedStates(object):
        Enabled = pywbem.Uint16(2)
        Disabled = pywbem.Uint16(3)
        Shut_Down = pywbem.Uint16(4)
        Offline = pywbem.Uint16(6)
        Test = pywbem.Uint16(7)
        Defer = pywbem.Uint16(8)
        Quiesce = pywbem.Uint16(9)
        Reboot = pywbem.Uint16(10)
        Reset = pywbem.Uint16(11)
        # DMTF_Reserved = ..
        _reverse_map = {
                2: 'Enabled',
                3: 'Disabled',
                4: 'Shut Down',
                6: 'Offline',
                7: 'Test',
                8: 'Defer',
                9: 'Quiesce',
                10: 'Reboot',
                11: 'Reset'
        }

    class Status(object):
        OK = 'OK'
        Error = 'Error'
        Degraded = 'Degraded'
        Unknown = 'Unknown'
        Pred_Fail = 'Pred Fail'
        Starting = 'Starting'
        Stopping = 'Stopping'
        Service = 'Service'
        Stressed = 'Stressed'
        NonRecover = 'NonRecover'
        No_Contact = 'No Contact'
        Lost_Comm = 'Lost Comm'
        Stopped = 'Stopped'

    class CommunicationStatus(object):
        Unknown = pywbem.Uint16(0)
        Not_Available = pywbem.Uint16(1)
        Communication_OK = pywbem.Uint16(2)
        Lost_Communication = pywbem.Uint16(3)
        No_Contact = pywbem.Uint16(4)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'Not Available',
                2: 'Communication OK',
                3: 'Lost Communication',
                4: 'No Contact'
        }

    class OperationalStatus(object):
        Unknown = pywbem.Uint16(0)
        Other = pywbem.Uint16(1)
        OK = pywbem.Uint16(2)
        Degraded = pywbem.Uint16(3)
        Stressed = pywbem.Uint16(4)
        Predictive_Failure = pywbem.Uint16(5)
        Error = pywbem.Uint16(6)
        Non_Recoverable_Error = pywbem.Uint16(7)
        Starting = pywbem.Uint16(8)
        Stopping = pywbem.Uint16(9)
        Stopped = pywbem.Uint16(10)
        In_Service = pywbem.Uint16(11)
        No_Contact = pywbem.Uint16(12)
        Lost_Communication = pywbem.Uint16(13)
        Aborted = pywbem.Uint16(14)
        Dormant = pywbem.Uint16(15)
        Supporting_Entity_in_Error = pywbem.Uint16(16)
        Completed = pywbem.Uint16(17)
        Power_Mode = pywbem.Uint16(18)
        Relocating = pywbem.Uint16(19)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'Other',
                2: 'OK',
                3: 'Degraded',
                4: 'Stressed',
                5: 'Predictive Failure',
                6: 'Error',
                7: 'Non-Recoverable Error',
                8: 'Starting',
                9: 'Stopping',
                10: 'Stopped',
                11: 'In Service',
                12: 'No Contact',
                13: 'Lost Communication',
                14: 'Aborted',
                15: 'Dormant',
                16: 'Supporting Entity in Error',
                17: 'Completed',
                18: 'Power Mode',
                19: 'Relocating'
        }

    class OperatingStatus(object):
        Unknown = pywbem.Uint16(0)
        Not_Available = pywbem.Uint16(1)
        Servicing = pywbem.Uint16(2)
        Starting = pywbem.Uint16(3)
        Stopping = pywbem.Uint16(4)
        Stopped = pywbem.Uint16(5)
        Aborted = pywbem.Uint16(6)
        Dormant = pywbem.Uint16(7)
        Completed = pywbem.Uint16(8)
        Migrating = pywbem.Uint16(9)
        Emigrating = pywbem.Uint16(10)
        Immigrating = pywbem.Uint16(11)
        Snapshotting = pywbem.Uint16(12)
        Shutting_Down = pywbem.Uint16(13)
        In_Test = pywbem.Uint16(14)
        Transitioning = pywbem.Uint16(15)
        In_Service = pywbem.Uint16(16)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'Not Available',
                2: 'Servicing',
                3: 'Starting',
                4: 'Stopping',
                5: 'Stopped',
                6: 'Aborted',
                7: 'Dormant',
                8: 'Completed',
                9: 'Migrating',
                10: 'Emigrating',
                11: 'Immigrating',
                12: 'Snapshotting',
                13: 'Shutting Down',
                14: 'In Test',
                15: 'Transitioning',
                16: 'In Service'
        }

    class RequestStateChange(object):
        Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unknown_or_Unspecified_Error = pywbem.Uint32(2)
        Cannot_complete_within_Timeout_Period = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        In_Use = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        Method_Parameters_Checked___Job_Started = pywbem.Uint32(4096)
        Invalid_State_Transition = pywbem.Uint32(4097)
        Use_of_Timeout_Parameter_Not_Supported = pywbem.Uint32(4098)
        Busy = pywbem.Uint32(4099)
        # Method_Reserved = 4100..32767
        # Vendor_Specific = 32768..65535
        class RequestedState(object):
            Enabled = pywbem.Uint16(2)
            Disabled = pywbem.Uint16(3)
            Shut_Down = pywbem.Uint16(4)
            Offline = pywbem.Uint16(6)
            Test = pywbem.Uint16(7)
            Defer = pywbem.Uint16(8)
            Quiesce = pywbem.Uint16(9)
            Reboot = pywbem.Uint16(10)
            Reset = pywbem.Uint16(11)
            # DMTF_Reserved = ..
            # Vendor_Reserved = 32768..65535

    class InstallFromByteStream(object):
        Job_Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unspecified_Error = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Target_In_Use = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        Method_Parameters_Checked___Job_Started = pywbem.Uint32(4096)
        Unsupported_TargetType = pywbem.Uint32(4097)
        Unattended_silent_installation_not_supported = pywbem.Uint32(4098)
        Downgrade_reinstall_not_supported = pywbem.Uint32(4099)
        Not_enough_memory = pywbem.Uint32(4100)
        Not_enough_swap_space = pywbem.Uint32(4101)
        Unsupported_version_transition = pywbem.Uint32(4102)
        Not_enough_disk_space = pywbem.Uint32(4103)
        Software_and_target_operating_system_mismatch = pywbem.Uint32(4104)
        Missing_dependencies = pywbem.Uint32(4105)
        Not_applicable_to_target = pywbem.Uint32(4106)
        No_supported_path_to_image = pywbem.Uint32(4107)
        # Method_Reserved = 4108..32767
        # Vendor_Specific = 32768..65535

    InstallFromByteStream.InstallOptions = InstallFromURI.InstallOptions

    class InstallFromSoftwareIdentity(object):
        Job_Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unspecified_Error = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Target_In_Use = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        Method_Parameters_Checked___Job_Started = pywbem.Uint32(4096)
        Unsupported_TargetType = pywbem.Uint32(4097)
        Unattended_silent_installation_not_supported = pywbem.Uint32(4098)
        Downgrade_reinstall_not_supported = pywbem.Uint32(4099)
        Not_enough_memory = pywbem.Uint32(4100)
        Not_enough_swap_space = pywbem.Uint32(4101)
        Unsupported_version_transition = pywbem.Uint32(4102)
        Not_enough_disk_space = pywbem.Uint32(4103)
        Software_and_target_operating_system_mismatch = pywbem.Uint32(4104)
        Missing_dependencies = pywbem.Uint32(4105)
        Not_applicable_to_target = pywbem.Uint32(4106)
        No_supported_path_to_image = pywbem.Uint32(4107)
        Cannot_add_to_Collection = pywbem.Uint32(4108)
        # Method_Reserved = 4109..32767
        # Vendor_Specific = 32768..65535

    InstallFromSoftwareIdentity.InstallOptions = InstallFromURI.InstallOptions

    class StartMode(object):
        Automatic = 'Automatic'
        Manual = 'Manual'

    class PrimaryStatus(object):
        Unknown = pywbem.Uint16(0)
        OK = pywbem.Uint16(1)
        Degraded = pywbem.Uint16(2)
        Error = pywbem.Uint16(3)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'OK',
                2: 'Degraded',
                3: 'Error'
        }

    class FindIdentity(object):
        Found = pywbem.Uint32(0)
        NoMatch = pywbem.Uint32(1)

def get_path(env):
    """@return instance name with prefilled properties"""
    systemop = ComputerSystem.get_path(env)
    clsname = "LMI_SoftwareInstallationService"
    op = util.new_instance_name(clsname,
            CreationClassName=clsname,
            SystemCreationClassName=systemop.classname,
            SystemName=systemop["Name"],
            Name="LMI:LMI_SoftwareInstallationService")
    return op

@cmpi_logging.trace_function
def check_path(env, service, prop_name):
    """
    Checks instance name of SoftwareInstallationService.
    @param prop_name name of object name; used for error descriptions
    """
    if not isinstance(service, pywbem.CIMInstanceName):
        raise pywbem.CIMError(pywbem.CIM_ERR_INVALID_PARAMETER,
                "\"%s\" must be a CIMInstanceName" % prop_name)
    our_service = get_path(env)
    ch = env.get_cimom_handle()
    if service.namespace != our_service.namespace:
        raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                'Namespace of "%s" does not match "%s"' % (
                    prop_name, our_service.namespace))
    if not ch.is_subclass(our_service.namespace,
            sub=service.classname,
            super=our_service.classname):
        raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                "Class of \"%s\" must be a subclass of %s" % (
                    prop_name, our_service.classname))
    for key in our_service.keybindings.keys():
        if not key in service:
            raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                    "\"%s\" is missing %s key property" % ( prop_name, key))
        if key == "SystemName":
            continue
    if not is_this_system(service["SystemName"]):
        raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
                "\"%s\" key property SystemName(%s) does not match \"%s\"" % (
                prop_name, service[key], our_service[key]))
    return True

@cmpi_logging.trace_function
def check_path_property(env, op, prop_name):
    """
    Checks, whether prop_name property of op object path is correct.
    If not, an exception will be raised.
    """
    if not prop_name in op:
        raise pywbem.CIMError(pywbem.CIM_ERR_INVALID_PARAMETER,
                "Missing %s key property!" % prop_name)
    return check_path(env, op[prop_name], prop_name)

@cmpi_logging.trace_function
def _check_target_and_collection(env, method, target, collection):
    """
    Checks Target and Collection parameters of provider's installation
    methods.
    """
    values = Values.InstallFromSoftwareIdentity
    if target:
        try:
            ComputerSystem.check_path(env, target, "Target")
        except pywbem.CIMError as exc:
            raise InstallationError(values.Unspecified_Error,
                    "Target must be either NULL or match managed"
                    " computer system: %s", str(exc))
    if collection:
        try:
            SystemCollection.check_path(env, collection, "Collection")
        except pywbem.CIMError as exc:
            raise InstallationError(values.Unspecified_Error,
                    "Collection does not match system software collection: %s" %
                    str(exc))
    if target and collection:
        raise InstallationError(values.Unspecified_Error,
                "Only one of Target and Collection parameters can be specified"
                " at the same time.")
    if not target and not collection:
        raise InstallationError(values.Unspecified_Error,
                  "Either Target or Collection parameter must be specified."
             if method == Job.JOB_METHOD_INSTALL_FROM_SOFTWARE_IDENTITY
             else "Missing Target parameter.")

@cmpi_logging.trace_function
def _install_or_remove_check_params(
        env, method, source, target, collection,
        install_options,
        install_options_values):
    """
    Checks parameters of provider's installation methods.
    Upon any invalid option an InstallationError will be raised.
    @return tuple (action, force, repair)
      where action is one of Values.InstallFromSoftwareIdentity properties
    """
    if not method in (
            Job.JOB_METHOD_INSTALL_FROM_URI,
            Job.JOB_METHOD_INSTALL_FROM_SOFTWARE_IDENTITY,
            Job.JOB_METHOD_INSTALL_FROM_BYTE_STREAM):
        raise ValueError("unknown method")

    values = Values.InstallFromSoftwareIdentity
    supported_options = values.InstallOptions.supported.copy()
    if method == Job.JOB_METHOD_INSTALL_FROM_URI:
        supported_options.remove(values.InstallOptions.Uninstall)

    if not source:
        raise InstallationError(values.Unspecified_Error,
                "Missing %s parameter." % (JOB_METHOD_SRC_PARAM_NAMES[method]))
    if not install_options:
        install_options = []
    elif not isinstance(install_options, list):
        raise InstallationError(values.Unspecified_Error,
                "InstallOptions must be a list of uint16 values.")
    options = {p for p in install_options}

    if options - supported_options:
        raise InstallationError(values.Unspecified_Error,
                "unsupported install options: {%s}" %
                ", ".join([   values.InstallOptions._reverse_map[p]
                          for p in options - supported_options]))
    if install_options_values and len(options) != len(install_options_values):
        raise InstallationError(values.Unspecified_Error,
                "InstallOptions array must have the same"
                " length as InstallOptionsValues: %d != %d" % (
                    len(options), len(install_options_values)))
    if install_options_values:
        for opt, val in zip(install_options, install_options_values):
            if val:
                raise InstallationError(values.Unspecified_Error,
                        "install option \"%s\" can not have any"
                        " associated value: %s" % (opt, val))
    _check_target_and_collection(env, method, target, collection)
    exclusive = [opt for opt in options if opt in {
        values.InstallOptions.Install,
        values.InstallOptions.Update,
        values.InstallOptions.Uninstall }]
    if len(exclusive) > 1:
        raise InstallationError(values.Unspecified_Error,
            "specified more than one mutually exclusive option at once: {%s}" %
            ", ".join([   values.InstallOptions._reverse_map[p]
                      for p in exclusive]))
    if not exclusive:
        exclusive.append(values.InstallOptions.Install)
    return ( exclusive[0]
           , values.InstallOptions.Force_installation in options
           , values.InstallOptions.Repair in options)

@cmpi_logging.trace_function
def make_job_input_params(method, source, target, collection,
        install_options=None,
        install_options_values=None):
    """
    Make dictionary of input parameters, that are stored in job's metadata.
    This dictionary is used in creation of CIM_ConcreteJob and
    CIM_InstMethodCall.
    """
    input_params = {
        "Target" : pywbem.CIMProperty(
            name="Target", type="reference", value=target),
    }

    if method == Job.JOB_METHOD_VERIFY_INSTALLED_IDENTITY:
        input_params["Source"] = pywbem.CIMProperty(
            name="Source", type="reference", value=source)
        return input_params

    input_params["InstallOptionsValues"] = pywbem.CIMProperty(
            name="InstallOptionsValues",
            type="string",
            is_array=True,
            value=install_options_values)
    input_params["InstallOptions"] = pywbem.CIMProperty(
            name="InstallOptions", type="uint16",
            is_array=True,
            value=install_options)
    if method == Job.JOB_METHOD_INSTALL_FROM_URI:
        input_params["URI"] = pywbem.CIMProperty(
            name="URI", type="string", value=source)
    elif method == Job.JOB_METHOD_INSTALL_FROM_SOFTWARE_IDENTITY:
        input_params["Source"] = pywbem.CIMProperty(
            name="Source", type="reference", value=source)
        input_params["Collection"] = pywbem.CIMProperty(
            name="Collection", type="reference", value=collection)
    elif method == Job.JOB_METHOD_INSTALL_FROM_BYTE_STREAM:
        input_params["Image"] = pywbem.CIMProperty(
            name="Image", type="reference", value=source)

    return input_params

@cmpi_logging.trace_function
def install_or_remove_package(env, method,
        source, target, collection,
        install_options,
        install_options_values):
    """
    :param method: (``int``) Identifier of method defined in ``core.Job``
        module with variables prefixed with ``JOB_METHOD_``.
    """
    values = Values.InstallFromSoftwareIdentity
    (action, force, repair) = _install_or_remove_check_params(
            env, method, source, target, collection,
            install_options, install_options_values)
    input_params = make_job_input_params(method,
            source, target, collection, install_options,
            install_options_values)
    metadata = { "method" : method, "input_params" : input_params }
    try:
        ydb = YumDB.get_instance()
        if action == values.InstallOptions.Uninstall:
            src = Identity.object_path2nevra(source, with_epoch='ALWAYS')
            LOG().info('removing package %s', src)
            jobid = ydb.remove_package(src, async=True, **metadata)
        else:
            update = action == values.InstallOptions.Update
            if method == Job.JOB_METHOD_INSTALL_FROM_URI:
                LOG().info('%s package "%s"',
                        'updating' if update else 'installing', source)
                src = source
                jobid = ydb.install_package_from_uri(
                        source, update_only=update, force=force or repair,
                        async=True, **metadata)
            else:   # software identity
                src = Identity.object_path2nevra(source, with_epoch='ALWAYS')
                if update:
                    jobid = ydb.update_package(src,
                            force=force or repair, async=True, **metadata)
                else:
                    jobid = ydb.install_package(src,
                            force=force or repair, async=True, **metadata)
        LOG().info('installation job %s for pkg "%s" enqueued', jobid, src)
        return jobid

    except (pywbem.CIMError, errors.InvalidURI) as exc:
        LOG().exception('failed to install/remove package "%s" from %s: %s',
                source, JOB_METHOD_SRC_PARAM_NAMES[method].lower(), str(exc))
        raise InstallationError(values.Unspecified_Error, str(exc))

@cmpi_logging.trace_function
def verify_package(env, method, source, target):
    """
    :param method: (``int``) Identifier of method defined in ``core.Job``
        module with variables prefixed with ``JOB_METHOD_``.
    """
    values = Values.VerifyInstalledIdentity

    if method != Job.JOB_METHOD_VERIFY_INSTALLED_IDENTITY:
        raise ValueError("unknown method")

    if not source:
        raise InstallationError(values.Unspecified_Error,
                "Missing %s parameter." % (JOB_METHOD_SRC_PARAM_NAMES[method]))

    _check_target_and_collection(env, method, target, None)

    input_params = make_job_input_params(method, source, target, None)
    metadata = { "method" : method, "input_params" : input_params }
    nevra = Identity.object_path2nevra(source, with_epoch='ALWAYS')
    ydb = YumDB.get_instance()
    LOG().info('verifying package "%s"', nevra)
    jobid = ydb.check_package(nevra, async=True, **metadata)
    LOG().info('verification job %s for pkg "%s" enqueued', jobid, nevra)
    return jobid




# Copyright (C) 2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Jan Safranek <jsafrane@redhat.com>
# -*- coding: utf-8 -*-
"""
Module with functionality to create timers, which can be used in CMPI providers.

Default python threading.Timer is not suitable, because it creates thread
for each timer, which is inefficient. In addition, each such thread would need
to be registered at CIMOM to enable logging in timer callbacks.

Usage:

1. Initialize the TimerManager when your provider initializes!
Otherwise you may encounter weird exceptions.

2. When any provider needs timer, create it using Time.create_timer() to create
Timer instance.

3. Call Timer.start() to start the timer. It will call registered callback
when the timer expires. The callback is called in context of TimerManager
thread, which has enabled logging to CIMOM, i.e. the callback can log as usual.

4. (optionally) cancel the timer before expiration using Timer.cancel().
However, this does not guarantee that the timer callback won't be called -
it may be already being scheduled / called.

.. autoclass:: TimerManager
        :members:

.. autoclass:: Timer
        :members:

.. autoclass:: MonotonicClock
        :members:
"""

import ctypes
import threading
import Queue

from lmi.base import singletonmixin
from lmi.providers import cmpi_logging

LOG = cmpi_logging.get_logger(__name__)

class TimerException(Exception):
    pass

class MonotonicClock(object):
    """
    Monotonic clock, represented by clock_gettime() and CLOCK_MONOTONIC.
    This clock is not influenced by NTP or administrator setting time or date.
    """
    CLOCK_MONOTONIC = ctypes.c_int(1)

    class timespec(ctypes.Structure):
        _fields_ = [
                ("tv_sec", ctypes.c_long),
                ("tv_nsec", ctypes.c_long)]

    def __init__(self):
        libc = ctypes.CDLL("librt.so.1")
        self._clock_gettime = libc.clock_gettime

    def now(self):
        """
        Return current time, i.e. float representing seconds with precision up
        to nanoseconds (depends on glibc). The actual value of current time is
        meaningless, it can be used only to measure time differences.

        :returns: ``float`` with current time in seconds.
        """
        t = MonotonicClock.timespec(0, 0)
        ret = self._clock_gettime(self.CLOCK_MONOTONIC, ctypes.pointer(t))

        if ret < 0:
            raise TimerException("Cannot get clock time, clock_gettime() failed.")
        return t.tv_sec + t.tv_nsec * 10 ** (-9)

class Timer(object):
    """
    A class representing a timer. A timer has a timeout and after the timeout,
    given callback is called and the timer is deleted.
    """

    @cmpi_logging.trace_method
    def __init__(self, timer_manager, name, callback=None, *args, **kwargs):
        """
        Create a timer. If specified, given callback is registered.
        The callback is called with *args and **kwargs.

        :param timer_manager: (``TimerManager)`` Instance of the timer manager
            which will manage the timer.
        :param name: (``string``) Name of the timer, used for logging.
        :param callback: (``function``) Callback to call when the timer expires.
        :param *args, **kwargs: Parameters of the callback.
        """
        self._mgr = timer_manager
        self._name = name
        self._callback = callback
        self._args = args
        self._kwargs = kwargs

        LOG().trace_info("Timer: Timer %s created", name)

    @cmpi_logging.trace_method
    def set_callback(self, callback, *args, **kwargs):
        """
        Set callback to call when the timer expires.

        :param callback: (``function``) Callback to call when the timer expires.
        :param *args, **kwargs: Parameters of the callback.
        """
        self._callback = callback
        self._args = args
        self._kwargs = kwargs

    @cmpi_logging.trace_method
    def start(self, timeout):
        """
        Start the timer with given timeout. After the timeout, the registered
        callback will be called.

        :param timeout: (``float``) Timeout in seconds.
        """

        self._timeout = timeout
        now = self._mgr.now()
        self._end_time = now + timeout
        LOG().trace_info("Timer: Timer %s started at %f for %f seconds",
                self._name, now, self._timeout)
        self._mgr._add_timer(self)

    @cmpi_logging.trace_method
    def cancel(self):
        """
        Cancel the timer. This method does not guarantee that the callback won't
        be called, the timer might be calling the callback right now,
        """
        LOG().trace_info("Timer: Timer %s cancelled", self._name)
        self._mgr._remove_timer(self)

    @cmpi_logging.trace_method
    def _expired(self, now):
        """
        Returns True, if the timer is expired.

        :param now: (``float``) Current time, as returned by MonotonicClock.now().
        :returns: (``boolean``) ``True``, if the timer is expired.
        """
        if self._end_time <= now:
            LOG().trace_info("Timer: Timer %s has expired", self._name)
            return True
        return False

    @cmpi_logging.trace_method
    def _expire(self):
        """
        Called when the timer expired. It calls the callback.
        """
        LOG().trace_info("Timer: Calling callback for timer %s", self._name)
        self._callback(*self._args, **self._kwargs)

class TimerManager(singletonmixin.Singleton):
    """
    Manages set of timers.

    Python standard Timer class creates a thread for

    each timer, which is inefficient. This class uses only one thread, which
    is registered at CIMOM, i.e. it can log as usual.

    This class is singleton, use TimerManager.get_instance() to get the
    instance.

    Still, the singleton needs to be initialized with ProviderEnvironment to
    enable logging in the timer thread. Use TimerManager.get_instance(env) in
    you provider initialization.
    """

    # Commands to the timer thread
    COMMAND_STOP = 1
    COMMAND_RESCHEDULE = 2

    @cmpi_logging.trace_method
    def __init__(self, env=None):
        """
        Initialize new thread manager.

        :param env: (``ProviderEnvironment``) Environment to use for logging.
        """
        self._clock = MonotonicClock()
        self._lock = threading.RLock()
        self._queue = Queue.Queue()

        # Array of timers. Assumption: nr. of timers is relatively small,
        # i.e. hundreds at the worst.
        self._timers = []

        new_broker = None
        if env:
            broker = env.get_cimom_handle()
            new_broker = broker.PrepareAttachThread()

        self._timer_thread = threading.Thread(
                target=self._timer_loop, args=(new_broker,))
        self._timer_thread.daemon = False
        self._timer_thread.start()

    def create_timer(self, name, callback=None, *args, **kwargs):
        """
        Create new timer. If specified, given callback is registered.
        The callback is called with *args and **kwargs.

        :param name: (``string``) Name of the timer, used for logging.
        :param callback: (``function``) Callback to call when the timer expires.
        :param *args, **kwargs: Parameters of the callback.
        """
        return Timer(self, name, callback, *args, **kwargs)

    def _timer_loop(self, broker):
        """
        TimerManager thread main loop. It waits for timeout of all timers
        and calls their callbacks.

        :param broker: (``BrokerCIMOMHandle``) CIM broker handle, used for
                logging.
        """
        if broker:
            broker.AttachThread()
        LOG().info("Started Timer thread.")
        while True:
            self._handle_expired()
            timeout = self._find_timeout()
            if timeout != 0:
                # Wait for the timeout or any change in timers.
                try:
                    command = self._queue.get(timeout=timeout)
                    self._queue.task_done()
                    if command == self.COMMAND_STOP:
                        break  # stop the thread
                    # process COMMAND_RESCHEDULE in next loop
                except Queue.Empty:
                    # Timeout has happened, ignore the exception.
                    pass
        LOG().info("Stopped Timer thread.")

    @cmpi_logging.trace_method
    def _handle_expired(self):
        """
        Finds all expired timers, calls their callback and removes them from
        list of timers.
        """

        # Get list of expired timers.
        with self._lock:
            now = self.now()
            LOG().trace_info("Timer: Checking for expired, now=%f.", now)
            expired = [t for t in self._timers if t._expired(now)]

        # Call the callbacks (unlocked!).
        for t in expired:
            t._expire()

        # Remove the timers (locked).
        with self._lock:
            for t in expired:
                try:
                    LOG().trace_info("Timer: Removing %s", t._name)
                    self._timers.remove(t)
                except ValueError:
                    # The timer has already been removed.
                    pass

    @cmpi_logging.trace_method
    def _find_timeout(self):
        """
        Return nearest timeout, in seconds (as float, i.e. subsecond timeout
        is possible). If no timer is scheduled, None is returned.
        If there are expired timers, 0 is returned.

        :returns: Positive ``float``: Nearest timeout.
        :returns: ``0``: Some timer has expired.
        :returns: ``None``: No timer is scheduled.
        """
        with self._lock:
            if not self._timers:
                LOG().trace_info("Timer: No timers scheduled, waiting forever.")
                return None
            closest = min(self._timers, key=lambda timer: timer._end_time)
            now = self.now()
            timeout = closest._end_time - now
            if timeout > 0:
                LOG().trace_info("Timer: Waiting for %f seconds, now=%f.",
                        timeout, now)
                return timeout
            LOG().trace_info(
                    "Timer: Some timer has already expired, no waiting.")
            return 0

    @cmpi_logging.trace_method
    def _add_timer(self, timer):
        """
        Adds timer to list of timers. The timer must be started, i.e. its
        timeout must be nozero!
        This is internal method called by Timer.start().

        :param timer: (``Timer``) Timer to add.
        """
        with self._lock:
            self._timers.append(timer)
        # Wake up the timer manager thread.
        self._queue.put(self.COMMAND_RESCHEDULE)
        LOG().trace_info("Timer: Timer %s added", timer._name)

    @cmpi_logging.trace_method
    def _remove_timer(self, timer):
        """
        Remove timer from list of timers.
        This is internal method called by Timer.cancel().
        :param timer: (``Timer``) Timer to remove.
        """
        with self._lock:
            try:
                self._timers.remove(timer)
            except ValueError:
                pass
        # Wake up the timer manager thread.
        self._queue.put(self.COMMAND_RESCHEDULE)
        LOG().trace_info("Timer: Timer %s removed", timer._name)

    def now(self):
        """
        Return current time, not influenced by NTP or admin setting date or
        time. The actual value of current time is meaningless, it can be used
        only to measure time differences.

        :returns: ``float`` Current time, in seconds.
        """
        return self._clock.now()

    @cmpi_logging.trace_method
    def shutdown(self):
        """
        Stop the thread. This method blocks until the thread is safely
        destroyed.
        """
        self._queue.put(self.COMMAND_STOP)
        self._timer_thread.join()

if __name__ == "__main__":
    LOG = cmpi_logging.CMPILogger("")
    import time

    class Env(object):
        def AttachThread(self):
            pass
        def PrepareAttachThread(self):
            return self
        def get_cimom_handle(self):
            return self

    clock = MonotonicClock()

    start = clock.now()
    time.sleep(0.5)
    print "Clock 0.5:", clock.now() - start

    time.sleep(0.5)
    print "Clock 1:", clock.now() - start

    mgr = TimerManager.get_instance(Env())

    def callback(msg):
        if callback.first:
            t = mgr.create_timer("internal 0.5")
            t.set_callback(callback, "internal 0.5")
            t.start(0.5)
            callback.first = False

        print clock.now(), msg

    callback.first = True

    t1 = mgr.create_timer("one second")
    t1.set_callback(callback, "1")
    t1.start(1)
    t2 = mgr.create_timer("two seconds")
    t2.set_callback(callback, "2")
    t2.start(2)
    t22 = mgr.create_timer("two seconds 2")
    t22.set_callback(callback, "2 again")
    t22.start(2)
    t15 = mgr.create_timer("one+half seconds")
    t15.set_callback(callback, "1.5")
    t15.start(1.5)

    time.sleep(4)

    mgr.stop_thread()



#!/usr/bin/env python
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Unit tests for ``LMI_ResourceForSoftwareIdentity`` provider.
"""

import unittest

import package
import swbase

ENABLED_STATE_ENABLED = 2

class TestResourceForSoftwareIdentity(swbase.SwTestCase):
    """
    Basic cim operations test.
    """

    CLASS_NAME = "LMI_ResourceForSoftwareIdentity"
    KEYS = ("ManagedElement", "AvailableSAP")

    def make_op(self, pkg=None, repo=None):
        """
        :returns: Object path of ``LMI_ResourceForSoftwareIdentity``
        :rtype: :py:class:`lmi.shell.LMIInstanceName`
        """
        objpath = self.cim_class.new_instance_name({
            "AvailableSAP" : self.ns.LMI_SoftwareIdentityResource. \
                new_instance_name({
                    "CreationClassName" : "LMI_SoftwareIdentityResource",
                    "SystemCreationClassName" : self.system_cs_name,
                    "SystemName" : self.SYSTEM_NAME
                }),
            "ManagedElement" :
                self.ns.LMI_SoftwareIdentity.new_instance_name({})
            })
        if repo is not None:
            objpath.AvailableSAP.wrapped_object["Name"] = repo.repoid
        elif pkg is not None:
            objpath.AvailableSAP.wrapped_object["Name"] = pkg.repoid
        if pkg is not None:
            objpath.ManagedElement.wrapped_object["InstanceID"] = (
                      'LMI:LMI_SoftwareIdentity:' + pkg.nevra)
        return objpath

    @swbase.test_with_repos('stable', 'updates', 'misc',
            **{'updates-testing' : True})
    def test_get_instance(self):
        """
        Test ``GetInstance()`` call on ``LMI_ResourceForSoftwareIdentity``.
        """
        for repo in self.repodb.values():
            self.assertTrue(repo.status)
            for pkg in repo.packages:
                objpath = self.make_op(pkg, repo)
                inst = objpath.to_instance()
                self.assertNotEqual(inst, None)
                self.assertEqual(set(inst.path.key_properties()),
                        set(self.KEYS))
                for key in self.KEYS:
                    self.assertCIMNameEqual(
                        getattr(inst, key), getattr(inst.path, key),
                        'Key property "%s" does not match for package "%s#%s"!'
                        % (key, repo.repoid, pkg))

    @swbase.test_with_repos(**{'updates' : False})
    def test_get_instance_disabled_repo(self):
        """
        Test ``GetInstance()`` call on ``LMI_ResourceForSoftwareIdentity``
        with disabled repository.
        """
        repo = self.get_repo('updates')
        self.assertFalse(repo.status)
        for pkg in repo.packages:
            objpath = self.make_op(pkg, repo)
            inst = objpath.to_instance()
            self.assertNotEqual(inst, None)
            self.assertEqual(set(inst.path.key_properties()),
                    set(self.KEYS))
            for key in self.KEYS:
                self.assertCIMNameEqual(
                    getattr(inst, key), getattr(inst.path, key),
                    'Key property "%s" does not match for package "%s#%s"!' %
                    (key, repo.repoid, pkg))

    @swbase.test_with_repos('stable')
    def test_repo_identity_names(self):
        """
        Test ``AssociatorNames()`` call on ``LMI_ResourceForSoftwareIdentity``.
        """
        repo = self.get_repo('stable')
        self.assertTrue(repo.status)
        self.assertGreater(repo.pkg_count, 0)

        objpath = self.make_op(repo=repo)
        refs = objpath.AvailableSAP.to_instance().associator_names(
                AssocClass=self.CLASS_NAME,
                Role="AvailableSAP",
                ResultRole="ManagedElement",
                ResultClass="LMI_SoftwareIdentity")
        self.assertEqual(len(refs), repo.pkg_count,
                'repository "%s" is missing software identities'
                % repo.name)
        for ref in refs:
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentity")
            self.assertEqual(ref.key_properties(), ["InstanceID"])
            self.assertTrue(
                    ref.InstanceID.startswith("LMI:LMI_SoftwareIdentity:"))

        nevra_set = set(i.InstanceID for i in refs)
        for pkg in repo.packages:
            nevra = 'LMI:LMI_SoftwareIdentity:'+pkg.nevra
            self.assertTrue(nevra in nevra_set,
                    'Missing nevra "%s" for repo "%s".' % (nevra,
                        repo.repoid))
            nevra_set.remove(nevra)
        self.assertEqual(len(nevra_set), 0,
                "all packages from repository have been listed")

    @swbase.test_with_repos(stable=False)
    def test_disabled_repo_identity_names(self):
        """
        Test ``AssociatorNames()`` call on ``LMI_ResourceForSoftwareIdentity``.
        """
        repo = self.get_repo('stable')
        self.assertFalse(repo.status)
        self.assertGreater(repo.pkg_count, 0)

        objpath = self.make_op(repo=repo)
        refs = objpath.AvailableSAP.to_instance().associator_names(
                AssocClass=self.CLASS_NAME,
                Role="AvailableSAP",
                ResultRole="ManagedElement",
                ResultClass="LMI_SoftwareIdentity")
        self.assertEqual(len(refs), repo.pkg_count,
                'repository "%s" is missing software identities'
                % repo.name)
        for ref in refs:
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentity")
            self.assertEqual(ref.key_properties(), ["InstanceID"])
            self.assertTrue(
                    ref.InstanceID.startswith("LMI:LMI_SoftwareIdentity:"))

    @swbase.test_with_repos('updates')
    @swbase.test_with_packages(**{
            'updates#pkg1' : True,
            'updates#pkg2' : True,
            'updates#pkg3' : False,
            'updates#pkg4' : False
    })
    def test_repo_identities(self):
        """
        Test ``Associators()`` call on ``LMI_ResourceForSoftwareIdentity``.
        """
        repo = self.get_repo('updates')
        self.assertTrue(repo.status)
        self.assertGreater(repo.pkg_count, 0)

        objpath = self.make_op(repo=repo)
        refs = objpath.AvailableSAP.to_instance().associators(
                AssocClass=self.CLASS_NAME,
                Role="AvailableSAP",
                ResultRole="ManagedElement",
                ResultClass="LMI_SoftwareIdentity")
        if repo.pkg_count:
            self.assertGreater(len(refs), 0,
                'no software identities associated to repo "%s"'
                % repo.repoid)
        for ref in refs:
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentity")
            self.assertEqual(ref.path.key_properties(), ["InstanceID"])

        nevra_dict = {i.ElementName: i for i in refs}
        installed_count = 0
        for pkg in repo.packages:
            self.assertTrue(pkg.nevra in nevra_dict,
                    'Missing package "%s" in repo "%s".' % (pkg, repo.repoid))
            if package.is_pkg_installed(pkg):
                self.assertNotEqual(nevra_dict[pkg.nevra].InstallDate, None,
                        "InstallDate property is set for installed and"
                        " available package %s" % pkg)
                installed_count += 1
            else:
                self.assertEqual(nevra_dict[pkg.nevra].InstallDate, None,
                        "InstallDate property is unset for available package"
                        " %s" % pkg)
            del nevra_dict[pkg.nevra]
        self.assertEqual(len(nevra_dict), 0,
                "all packages from repository have been listed")
        self.assertEqual(installed_count, 2)

    @swbase.test_with_repos('stable')
    def test_get_pkg_repository_name(self):
        """
        Try to get software identities associated with repository.
        """
        repo = self.get_repo('stable')
        self.assertTrue(repo.status)
        for pkg in repo.packages:
            objpath = self.make_op(pkg, repo)
            inst = objpath.ManagedElement.to_instance()
            self.assertNotEqual(inst, None,
                    "GetInstance() succeeds for package %s" % pkg)
            refs = inst.associator_names(
                    AssocClass=self.CLASS_NAME,
                    Role="ManagedElement",
                    ResultRole="AvailableSAP",
                    ResultClass="LMI_SoftwareIdentityResource")
            self.assertEqual(1, len(refs),
                    'No repo found for pkg "%s".' % pkg.nevra)
            ref = refs[0]
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentityResource")
            self.assertEqual(set(ref.key_properties()),
                    set(["SystemCreationClassName", "SystemName",
                        "Name", "CreationClassName"]))
            self.assertEqual(ref.Name, pkg.repoid,
                    'Repository name does not match for pkg "%s"'
                    % pkg.nevra)

    @swbase.test_with_repos('updates')
    @swbase.test_with_packages(**{
        'updates#pkg1' : True,
        'updates#pkg2' : True,
        'updates#pkg3' : False,
        'updates#pkg4' : False
    })
    def test_get_pkg_repository(self):
        """
        Try to get repository for installed and not installed package.
        """
        repo = self.get_repo('updates')
        self.assertTrue(repo.status)
        for pkg in repo.packages:
            objpath = self.make_op(pkg, repo)
            inst = objpath.ManagedElement
            self.assertNotEqual(inst, None,
                    "GetInstance() succeeds for package %s" % pkg)
            refs = inst.to_instance().associators(
                    AssocClass=self.CLASS_NAME,
                    Role="ManagedElement",
                    ResultRole="AvailableSAP",
                    ResultClass="LMI_SoftwareIdentityResource")
            self.assertEqual(1, len(refs),
                    'No repo found for pkg "%s".' % pkg.nevra)
            ref = refs[0]
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentityResource")
            self.assertEqual(set(ref.path.key_properties()),
                    set(["SystemCreationClassName", "SystemName",
                        "Name", "CreationClassName"]))
            self.assertEqual(ref.Name, pkg.repoid,
                    'Repository name does not match for pkg "%s"'
                    % pkg.nevra)
            self.assertEqual(ref.ElementName, pkg.repoid)
            self.assertEqual(ref.EnabledState, ENABLED_STATE_ENABLED)

def suite():
    """For unittest loaders."""
    return unittest.TestLoader().loadTestsFromTestCase(
            TestResourceForSoftwareIdentity)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Unit tests for ``LMI_MemberOfSoftwareCollection`` provider.
"""

import unittest

import swbase

class TestMemberOfSoftwareCollection(swbase.SwTestCase):
    """
    Basic cim operations test for ``LMI_MemberOfSoftwareCollection``.
    """

    CLASS_NAME = "LMI_MemberOfSoftwareCollection"
    KEYS = ("Collection", "Member")

    def make_op(self, pkg):
        """
        :returns Object path of ``LMI_MemberOfSoftwareCollection``.
        :rtype: :py:class:`lmi.shell.LMIInstanceName`
        """
        return self.cim_class.new_instance_name({
            "Collection" : self.ns.LMI_SystemSoftwareCollection. \
                new_instance_name({
                    "InstanceID" : "LMI:LMI_SystemSoftwareCollection"
                }),
            "Member" : self.ns.LMI_SoftwareIdentity.new_instance_name({
                "InstanceID" : 'LMI:LMI_SoftwareIdentity:' + pkg.nevra
                })
            })

    @swbase.test_with_repos('stable', 'misc', 'updates-testing')
    def test_get_instance(self):
        """
        Test ``GetInstance()`` call on ``LMI_MemberOfSoftwareCollection``.
        """
        for repoid in ('stable', 'misc', 'updates-testing'):
            repo = self.get_repo(repoid)
            for pkg in repo.packages:
                objpath = self.make_op(pkg)
                inst = objpath.to_instance()
                self.assertNotEqual(inst, None,
                        'Package "%s" is available in SystemSoftwareCollection'
                        % pkg)
                self.assertCIMNameEqual(inst.path, objpath,
                        "Object paths should match for package %s" % pkg)
                self.assertEqual(set(inst.path.key_properties()),
                        set(self.KEYS))

    @swbase.test_with_repos(misc=False)
    def test_get_instance_with_disabled_repo(self):
        """
        Test ``GetInstance()`` call on ``LMI_MemberOfSoftwareCollection``
        with package from disabled repository.
        """
        repo = self.get_repo('misc')
        for pkg in repo.packages:
            objpath = self.make_op(pkg)
            inst = objpath.to_instance()
            self.assertEqual(inst, None,
                'package should not be accessible from disabled repository.')

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates': False,
        'updates-testing': True,
        'misc' : False
    })
    @swbase.test_with_packages(**{
        'stable#pkg1' : True,
        'updates#pkg2' : True,
        'pkg3' : False,
        'pkg4' : False
    })
    def test_enum_instances(self):
        """
        Test ``EnumInstances()`` call on ``LMI_MemberOfSoftwareCollection``.
        """
        insts = self.cim_class.instances()
        self.assertGreater(len(insts), 0)
        for inst in insts:
            self.assertEqual(set(inst.path.key_properties()), set(self.KEYS))
            for key in self.KEYS:
                self.assertCIMNameEqual(
                        getattr(inst, key), getattr(inst.path, key))
        nevra_set = set(i.Member.InstanceID[
                len('LMI:LMI_SoftwareIdentity:'):] for i in insts)
        for repo in self.repodb.values():
            if not repo.repoid in (
                    'openlmi-sw-test-repo-stable',
                    'openlmi-sw-test-repo-updates-testing'):
                continue
            for pkg in repo.packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates': False,
        'updates-testing': True,
        'misc' : False
    })
    def test_enum_instance_names(self):
        """
        Test ``EnumInstanceNames()`` call on ``LMI_MemberOfSoftwareCollection``.
        """
        inames = self.cim_class.instance_names()
        self.assertGreater(len(inames), 0)
        objpath = self.make_op(self.get_repo('stable')['pkg1'])
        for iname in inames:
            self.assertEqual(iname.namespace, 'root/cimv2')
            self.assertEqual(iname.classname, self.CLASS_NAME)
            self.assertEqual(sorted(iname.key_properties()), sorted(self.KEYS))
            self.assertCIMNameEqual(objpath.Collection, iname.Collection)
        nevra_set = set(i.Member.InstanceID[
                len('LMI:LMI_SoftwareIdentity:'):] for i in inames)
        for repo in self.repodb.values():
            if not repo.repoid in (
                    'openlmi-sw-test-repo-stable',
                    'openlmi-sw-test-repo-updates-testing'):
                continue
            for pkg in repo.packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)


    @swbase.test_with_repos('stable')
    def test_get_package_collections(self):
        """
        Try to get instance of software collection associated with package.
        """
        repo = self.get_repo('stable')
        for pkg in repo.packages:
            objpath = self.make_op(pkg)
            inst = objpath.Member.to_instance()
            self.assertNotEqual(inst, None,
                    "package %s is available" % pkg)
            refs = inst.associators(
                    AssocClass=self.CLASS_NAME,
                    Role="Member",
                    ResultRole="Collection",
                    ResultClass="LMI_SystemSoftwareCollection")
            self.assertEqual(len(refs), 1,
                "exactly one collection is associated with package %s" % pkg)
            self.assertCIMNameEqual(refs[0].path, objpath.Collection)
            self.assertEqual(refs[0].InstanceID, objpath.Collection.InstanceID)

    @swbase.test_with_repos('stable')
    def test_get_package_collection_names(self):
        """
        Try to get instance of software collection associated with package.
        """
        repo = self.get_repo('stable')
        for pkg in repo.packages:
            objpath = self.make_op(pkg)
            inst = objpath.Member.to_instance()
            self.assertNotEqual(inst, None,
                    "package %s is available" % pkg)
            refs = inst.associator_names(
                    AssocClass=self.CLASS_NAME,
                    Role="Member",
                    ResultRole="Collection",
                    ResultClass="LMI_SystemSoftwareCollection")
            self.assertEqual(len(refs), 1,
                "exactly one collection is associated with package %s" % pkg)
            self.assertCIMNameEqual(refs[0], objpath.Collection)

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates' : False,
        'updates-testing' : True,
        'misc' : False
    })
    def test_get_collection_package_names(self):
        """
        Try to get package names associated to software collection.
        """
        pkg = self.get_repo('stable')['pkg1']
        objpath = self.make_op(pkg)
        collection = objpath.Collection.to_instance()
        self.assertNotEqual(collection, None,
                "GetInstance call succeeds on collection")
        refs = collection.associator_names(
                AssocClass=self.CLASS_NAME,
                Role="Collection",
                ResultRole="Member",
                ResultClass="LMI_SoftwareIdentity")
        self.assertGreater(len(refs), 0)

        for ref in refs:
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentity")
            self.assertEqual(ref.key_properties(), ["InstanceID"])
            self.assertTrue(ref.InstanceID.startswith(
                "LMI:LMI_SoftwareIdentity:"))

        nevra_set = set(   i.InstanceID[len('LMI:LMI_SoftwareIdentity:'):]
                       for i in refs)
        for repo in self.repodb.values():
            if not repo.repoid in (
                    'openlmi-sw-test-repo-stable',
                    'openlmi-sw-test-repo-updates-testing'):
                continue
            for pkg in repo.packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

    @swbase.test_with_repos(**{
        'stable' : True,
        'updates' : False,
        'updates-testing' : True,
        'misc' : False
    })
    def test_get_collection_packages(self):
        """
        Try to get packages associated to software collection.
        """
        pkg = self.get_repo('stable')['pkg1']
        objpath = self.make_op(pkg)
        collection = objpath.Collection.to_instance()
        self.assertNotEqual(collection, None,
                "GetInstance call succeeds on collection")
        refs = collection.associators(
                AssocClass=self.CLASS_NAME,
                Role="Collection",
                ResultRole="Member",
                ResultClass="LMI_SoftwareIdentity")
        self.assertGreater(len(refs), 0)

        for ref in refs:
            self.assertEqual(ref.namespace, 'root/cimv2')
            self.assertEqual(ref.classname, "LMI_SoftwareIdentity")
            self.assertEqual(ref.path.key_properties(), ["InstanceID"])
            self.assertTrue(ref.InstanceID.startswith(
                "LMI:LMI_SoftwareIdentity:"))

        nevra_set = set(i.ElementName for i in refs)
        for repo in self.repodb.values():
            if not repo.repoid in (
                    'openlmi-sw-test-repo-stable',
                    'openlmi-sw-test-repo-updates-testing'):
                continue
            for pkg in repo.packages:
                self.assertIn(pkg.nevra, nevra_set)
                nevra_set.remove(pkg.nevra)
        self.assertEqual(len(nevra_set), 0)

def suite():
    """For unittest loaders."""
    return unittest.TestLoader().loadTestsFromTestCase(
            TestMemberOfSoftwareCollection)

if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2012 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Jan Safranek <jsafrane@redhat.com>
# Authors: Michal Minar <miminar@redhat.com>
# -*- coding: utf-8 -*-
"""
Module for BaseConfiguration class.

BaseConfiguration
--------------------

.. autoclass:: BaseConfiguration
    :members:

"""

import ConfigParser
import logging
import os
import socket

from lmi.base.singletonmixin import Singleton

def convert_value(section, option, convert_func, value):
    """
    Return result of application of ``convert_func`` on value.
    If the conversion failes, error is logged and ValueError is raised.

    :param section: (``str``) Section of configuration file. Used for
        error message.
    :param option: (``str``) Option of configuration file. Used for
        error message.
    :param convert_func: (``type``) Conversion function to apply on passed
        value.
    :param value: (``basestring``) Value to convert.
    """
    if not isinstance(value, basestring):
        raise TypeError("value must be a string")
    try:
        if convert_func is bool:
            return value.lower() in ('1', 'y', 'yes', 'on', 'true')
        if convert_func is str and isinstance(value, unicode):
            return value.encode('utf-8')
        if convert_func is unicode and isinstance(value, str):
            return value.decode('utf-8')
        return convert_func(value)
    except ValueError as exc:
        logging.getLogger(__name__).error(
                'failed to convert value of "[%s]%s: %s', section, option,
                exc)
        raise

class BaseConfiguration(Singleton):
    """
        OpenLMI configuration file. By default, it resides in
        /etc/openlmi/${provider_prefix}/${provider_prefix}.conf.

        There should be only one instance of this class.
    """

    #: Allow to access global configuration object also from lmi.providers
    #: subpackage where ``get_instance()`` can not be used.
    #: This variable shall be set in ``__init__()`` method.
    INSTANCE = None
    CONFIG_DIRECTORY_TEMPLATE_TOPLEVEL = '/etc/openlmi/'
    CONFIG_DIRECTORY_TEMPLATE_PROVIDER = '/etc/openlmi/%(provider_prefix)s/'
    CONFIG_FILE_PATH_TEMPLATE_TOPLEVEL = \
            CONFIG_DIRECTORY_TEMPLATE_TOPLEVEL + 'openlmi.conf'
    CONFIG_FILE_PATH_TEMPLATE_PROVIDER = \
            CONFIG_DIRECTORY_TEMPLATE_PROVIDER + '%(provider_prefix)s.conf'

    PERSISTENT_PATH_TEMPLATE = '/var/lib/openlmi-%(provider_prefix)s/'
    SETTINGS_DIR = 'settings/'

    DEFAULT_OPTIONS = {
        'Namespace'       : 'root/cimv2',
        'SystemClassName' : 'PG_ComputerSystem',
        # Default logging level
        "Level"           : "ERROR",
        'DebugBlivet'     : 'false',
        'Stderr'          : 'false',
    }

    @classmethod
    def provider_prefix(cls):
        """
        This is responsibility of a subclass.

        :rtype: (``string`) Prefix of providers in lowercase. For example
            configuration class for storage providers would return "storage".

        Result is used to construct configuration paths.
        """
        raise NotImplementedError

    @classmethod
    def default_options(cls):
        """ :rtype: (``dict``) Dictionary of default values. """
        return cls.DEFAULT_OPTIONS

    @classmethod
    def config_directory_toplevel(cls):
        """ Base directory with toplevel configuration settings. """
        return cls.CONFIG_DIRECTORY_TEMPLATE_TOPLEVEL

    @classmethod
    def config_directory_provider(cls):
        """ Base directory with provider specific configuration settings. """
        return cls.CONFIG_DIRECTORY_TEMPLATE_PROVIDER % {
                'provider_prefix' : cls.provider_prefix() }


    @classmethod
    def persistent_path(cls):
        """ Base directory with persistent settings. """
        return cls.PERSISTENT_PATH_TEMPLATE % {
                'provider_prefix': cls.provider_prefix() }

    @classmethod
    def config_file_path_toplevel(cls):
        """ File path of toplevel configuration file. """
        return cls.CONFIG_FILE_PATH_TEMPLATE_TOPLEVEL

    @classmethod
    def config_file_path_provider(cls):
        """ File path of provider specific configuration file. """
        return cls.CONFIG_FILE_PATH_TEMPLATE_PROVIDER % {
                'provider_prefix' : cls.provider_prefix() }


    @classmethod
    def mandatory_sections(cls):
        """
            Return list of sections, that must be present in configuration
            file. If not present, they will be created in memory.
        """
        return ['Log', 'CIM']

    def __init__(self):
        """ Initialize and load a configuration file."""
        self._listeners = set()
        self.config = ConfigParser.SafeConfigParser(
                defaults=self.default_options())
        self.load()
        BaseConfiguration.INSTANCE = self

    def add_listener(self, callback):
        """
            Add a callback, which will be called when configuration is updated.
            The callback will be called with instance of this class as
            parameter:
              callback(config)
        """
        self._listeners.add(callback)

    def remove_listener(self, callback):
        """
            Remove previously registered callback.
        """

        self._listeners.remove(callback)

    def _call_listeners(self):
        """
            Call all listeners that configuration has updated.
        """
        for callback in self._listeners:
            callback(self)

    def load(self):
        """
            Load configuration from config files. Provider specific options
            overrides toplevel openlmi configuration.
            The files do not need to exist.
        """
        self.config.read([self.config_file_path_toplevel(), self.config_file_path_provider()])
        for section in self.mandatory_sections():
            if not self.config.has_section(section):
                self.config.add_section(section)
        self._call_listeners()

    @property
    def namespace(self):
        """ Return namespace of OpenLMI provider. """
        return self.config.get('CIM', 'Namespace')

    @property
    def system_class_name(self):
        """ Return SystemClassName of OpenLMI provider. """
        return self.config.get('CIM', 'SystemClassName')

    @property
    def logging_level(self):
        """ Return name of logging level in lower case. """
        return self.config.get('Log', 'Level').lower()

    @property
    def stderr(self):
        """ Return True if logging to stderr is enabled. """
        return self.config.getboolean('Log', 'Stderr')

    def file_path(self, section, option):
        """
        Return absolute file path for requested option.
        Relative path is converted to absolute one with config's directory
        as a prefix.
        """
        path = self.config.get(section, option)
        if not os.path.isabs(path):
            path = os.path.join(self.config_directory(), path)
        return path

    def get_safe(self, section, option, convert=str, fallback=None,
            *args, **kwargs):
        """
        Get the configuration option value as specified type in a safe way.
        Value is searched in this order:
            config_file -> defaults_dict -> fallback

        :param section: (``str``) Section name of option.
        :param option: (``str``) Option name.
        :param convert: (``type``) Is a conversion function for obtained
            value. If the value could not be converted, error message is
            generated and ``fallback`` is returned. This function is not
            applied to ``fallback`` value. Supported values are:
                str, unicode, int ,float, long, bool

        :param fallback: Value returned, when section or option does not
            exists and no default value is given, or when the obtained value
            could not be converted by supplied function.

        All the other parameters are passed to the ``SafeConfigParser.get()``
        method.
        """
        if not isinstance(section, basestring):
            raise TypeError('section must be a string')
        if not isinstance(option, basestring):
            raise TypeError("option must be a string")
        if not convert in (str, unicode, int, float, long, bool):
            raise ValueError("unsupported type for conversion: %s:",
                    getattr(convert, '__name__', 'unknown'))
        if (   not self.config.has_option(section, option)
           and not option.lower() in self.default_options()):
            logging.getLogger(__name__).warn(
                    'no option value and no default supplied for "[%s]%s"',
                    section, option)
            return fallback
        try:
            value = self.config.get(section, option, *args, **kwargs)
        except ConfigParser.Error as exc:
            logging.getLogger(__name__).error(
                    'failed to get value of "[%s]%s": %s', section, option,
                    exc)
            return fallback
        try:
            # first try to convert value from config
            return convert_value(section, option, convert, value)
        except ValueError as exc:
            logging.getLogger(__name__).error(
                    'failed to convert value of "[%s]%s: %s', section, option,
                    exc)
            # if it failes, try the value from defaults
            if (   option.lower() in self.default_options()
               and self.default_options()[option.lower()] != value):
                try:
                    return convert_value(section, option, convert,
                            self.default_options()[option.lower()])
                except ValueError:
                    pass    # error is already logged, no more options left
            return fallback



# -*- encoding: utf-8 -*-
# Software Management Providers
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Michal Minar <miminar@redhat.com>
#
"""
Define job classes representing kinds of jobs of worker process.
"""

import logging
import os
import threading
import time
import yum

from lmi.software import util
from lmi.software.yumdb import errors
from lmi.software.yumdb.packageinfo import PackageInfo
from lmi.software.yumdb.repository import Repository

class YumJob(object):  #pylint: disable=R0903
    """
    Base class for any job, that is processable by YumWorker process.
    It contains jobid attribute, that must be unique for
    each job, it's counted from zero a incremented after each creation.

    metadata attribute typically contain:
      name        - name of job, that is modifiable by user
      method      - identificator of method, that lead to creation of job
    """
    __slots__ = ( 'jobid', 'created', 'started', 'finished', 'last_change'
                , 'priority', 'result', 'result_data')

    # jobs can be created concurrently from multiple threads, that's
    # why we need to make its creation thread safe
    _JOB_ID_LOCK = threading.Lock()
    _JOB_ID = 0

    # job state enumeration
    NEW, RUNNING, COMPLETED, TERMINATED, EXCEPTION = range(5)
    # job result enumeration
    RESULT_SUCCESS, RESULT_TERMINATED, RESULT_ERROR = range(3)

    ResultNames = ("success", "terminated", "error")

    @staticmethod
    def _get_job_id():
        """
        Generates new job ids. It should be called only from constructor
        of YumJob. Ensures, that each job has a unique number.
        @return number of jobs created since program start -1
        """
        with YumJob._JOB_ID_LOCK:
            val = YumJob._JOB_ID
            YumJob._JOB_ID += 1
        return val

    @classmethod
    def handle_ignore_job_props(cls):
        """
        @return set of job properties, that does not count as job's handler
        arguments - job handler does not care fore metadata, jobid, priority,
        etc...
        """
        return set(YumJob.__slots__)

    def __init__(self, priority=None):
        if priority is None:
            priority = util.Configuration.get_instance().get_safe(
                    'Jobs', 'DefaultPriority', int)
        if not isinstance(priority, (int, long)):
            raise TypeError("priority must be integer")
        self.jobid = self._get_job_id()
        self.started = None
        self.finished = None
        self.priority = priority
        self.created = time.time()
        self.last_change = self.created
        self.result = None
        self.result_data = None

    @property
    def state(self):
        """
        @return integer representing job's state
        """
        if not self.started:
            return self.NEW
        if not self.finished:
            return self.RUNNING
        if self.result == self.RESULT_ERROR:
            return self.EXCEPTION
        if self.result == self.RESULT_TERMINATED:
            return self.TERMINATED
        return self.COMPLETED

    @property
    def job_kwargs(self):
        """
        Jobs are in worker handled in handlers specific for each subclass.
        These handlers are methods of worker. They accepts concrete arguments
        that can be obtained from job by invoking this property.
        @return dictionary of keyword arguments of job
        """
        kwargs = {}
        cls = self.__class__
        while not cls in (YumJob, object):
            for slot in cls.__slots__:
                if (   not slot in kwargs
                   and not slot in cls.handle_ignore_job_props()):
                    kwargs[slot] = getattr(self, slot)
            cls = cls.__bases__[0]
        for prop in YumJob.__slots__:
            kwargs.pop(prop, None)
        return kwargs

    def start(self):
        """Modify the state of job to RUNNING."""
        if self.started:
            raise errors.InvalidJobState("can not start already started job")
        self.started = time.time()
        self.last_change = self.started

    def finish(self, result, data=None):
        """
        Modify the state of job to one of {COMPLETED, EXCEPTION, TERMINATED}.
        Depending on result parameter.
        """
        if not self.started and result != self.RESULT_TERMINATED:
            raise errors.InvalidJobState("can not finish not started job")
        self.finished = time.time()
        if result == self.RESULT_TERMINATED:
            self.started = self.finished
        self.result = result
        self.result_data = data
        self.last_change = self.finished

    def update(self, **kwargs):
        """Change job's properties."""
        change = False
        for key, value in kwargs.items():
            if getattr(self, key) != value:
                setattr(self, key, value)
                change = True
        if change is True:
            self.last_change = time.time()

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.jobid == other.jobid

    def __ne__(self, other):
        return (  self.__class__ is not other.__class__
               or self.jobid != other.jobid)

    def __lt__(self, other):
        """
        JobControl jobs have the highest priority.
        """
        return (   other is not None    # terminating command
               and (  (  isinstance(self, YumJobControl)
                      and not isinstance(other, YumJobControl))
                   or (  self.priority < other.priority
                      or (   self.priority == other.priority
                         and (  self.jobid < other.jobid
                             or (   self.jobid == other.jobid
                                and (self.created < other.created)))))))

    def __cmp__(self, other):
        if other is None:   # terminating command
            return 1
        if (   isinstance(self, YumJobControl)
           and not isinstance(other, YumJobControl)):
            return -1
        if (   not isinstance(self, YumJobControl)
           and isinstance(other, YumJobControl)):
            return 1
        if self.priority < other.priority:
            return -1
        if self.priority > other.priority:
            return 1
        if self.jobid < other.jobid:
            return -1
        if self.jobid > other.jobid:
            return 1
        if self.created < other.created:
            return -1
        if self.created > other.created:
            return 1
        return 0

    def __str__(self):
        return "%s(id=%d,p=%d)" % (
                self.__class__.__name__, self.jobid, self.priority)

    def __getstate__(self):
        ret = self.job_kwargs
        for prop in self.handle_ignore_job_props():
            ret[prop] = getattr(self, prop)
        return ret

    def __setstate__(self, state):
        for k, value in state.items():
            setattr(self, k, value)

class YumAsyncJob(YumJob):              #pylint: disable=R0903
    """
    Base class for jobs, that support asynchronnous execution.
    No reply is sent upon job completition or error. The results are
    kept on server.
    """
    __slots__ = ( 'async'
                , 'delete_on_completion'
                , 'time_before_removal'
                , 'metadata')

    @classmethod
    def handle_ignore_job_props(cls):
        return YumJob.handle_ignore_job_props().union(YumAsyncJob.__slots__)

    def __init__(self, priority=None, async=False, metadata=None):
        YumJob.__init__(self, priority)
        self.async = bool(async)
        self.delete_on_completion = True
        self.time_before_removal = util.Configuration.get_instance().get_safe(
                'Jobs', 'DefaultTimeBeforeRemoval', float)
        if metadata is None and self.async is True:
            metadata = {}
        self.metadata = metadata

    def __str__(self):
        return "%s(id=%d,p=%d%s%s)" % (
                self.__class__.__name__, self.jobid,
                self.priority,
                ',async' if self.async else '',
                (',name="%s"'%self.metadata['name'])
                    if self.metadata and 'name' in self.metadata else '')

    def update(self, **kwargs):
        if 'metadata' in kwargs:
            self.metadata.update(kwargs.pop('metadata'))
        return YumJob.update(self, **kwargs)

# *****************************************************************************
# Job control funtions
# *****************************************************************************
class YumJobControl(YumJob):            #pylint: disable=R0903
    """Base class for any job used for asynchronous jobs management."""
    pass

class YumJobGetList(YumJobControl):     #pylint: disable=R0903
    """Request for obtaining list of all asynchronous jobs."""
    pass

class YumJobOnJob(YumJobControl):
    """
    Base class for any control job acting upon particular asynchronous job.
    """
    __slots__ = ('target', )
    def __init__(self, target):
        YumJobControl.__init__(self)
        if not isinstance(target, (int, long)):
            raise TypeError("target must be an integer")
        self.target = target

class YumJobGet(YumJobOnJob):         #pylint: disable=R0903
    """Get job object by its id."""
    pass

class YumJobGetByName(YumJobOnJob):   #pylint: disable=R0903
    """Get job object by its name property."""
    def __init__(self, name):
        YumJobOnJob.__init__(self, -1)
        self.target = name

class YumJobSetPriority(YumJobOnJob): #pylint: disable=R0903
    """Change priority of job."""
    __slots__ = ('new_priority', )

    def __init__(self, target, priority):
        YumJobOnJob.__init__(self, target)
        self.new_priority = priority

class YumJobUpdate(YumJobOnJob):      #pylint: disable=R0903
    """
    .. _YumJobUpdate:

    Update job's metadata. There are some forbidden properties, that
    can not be changed in this way. Those are all affecting job's priority
    and its scheduling for deletion. Plus any that store job's state.
    All forbidden properties are listed in ``FORBIDDEN_PROPERTIES``.
    """
    __slots__ = ('data', )
    FORBIDDEN_PROPERTIES = (
        'async', 'jobid', 'created', 'started', 'priority', 'finished',
        'delete_on_completion', 'time_before_removal', 'last_change')

    def __init__(self, target, **kwargs):
        YumJobOnJob.__init__(self, target)
        assert not set.intersection(
                set(YumJobUpdate.FORBIDDEN_PROPERTIES), set(kwargs))
        self.data = kwargs

class YumJobReschedule(YumJobOnJob):  #pylint: disable=R0903
    """Change the schedule of job's deletion."""
    __slots__ = ('delete_on_completion', 'time_before_removal')
    def __init__(self, target, delete_on_completion, time_before_removal):
        YumJobOnJob.__init__(self, target)
        if not isinstance(time_before_removal, (int, long, float)):
            raise TypeError("time_before_removal must be float")
        self.delete_on_completion = bool(delete_on_completion)
        self.time_before_removal = time_before_removal

class YumJobDelete(YumJobOnJob):      #pylint: disable=R0903
    """Delete job - can only be called on finished job."""
    pass

class YumJobTerminate(YumJobOnJob):   #pylint: disable=R0903
    """
    Can only be called on not yet started job.
    Running job can not be terminated.
    """
    pass

# *****************************************************************************
# Yum API functions
# *****************************************************************************
class YumLock(YumJob):              #pylint: disable=R0903
    """
    This job shall be sent when session starts. Session consists of several
    requests to ``YumWorker`` during which no other application shall
    interfere. Exclusive access to yum database is enforced with this job. When
    the session is over, YumUnlock job needs to be sent to YumWorker.
    """
    pass

class YumUnlock(YumJob):            #pylint: disable=R0903
    """
    This job shall be sent when session ends.
    """
    pass

class YumShutDown(YumJob):
    """
    Last job executed before provider terminates. It shuts down *YumWorker*,
    *SessionManager* and *JobManager*.
    """
    pass

class YumGetPackageList(YumJob):  #pylint: disable=R0903
    """
    Job requesing a list of packages.
    Arguments:
      kind - supported values are in SUPPORTED_KINDS tuple
        * installed lists all installed packages; more packages with
          the same name can be installed varying in their architecture
        * avail_notinst lists all available, not installed packages;
          allow_duplicates must be True to include older packages (but still
          available)
        * avail_reinst lists all installed packages, that are available;
          package can be installed, but not available anymore due to updates
          of repository, where only the newest packages are kept
        * available lists a union of avail_notinst and avail_reinst
        * all lists union of installed and avail_notinst

      allow_duplicates - whether multiple packages can be present
        in result for single (name, arch) of package differing
        in their version

      sort - whether to sort packages by nevra

      include_repos - either a string passable to RepoStorage.enableRepo()
        or a list of repository names, that will be temporared enabled before
        listing packages; this is applied after disabling of repositories

      exclude_repos - either a string passable to RepoStorage.disableRepo()
        or a list of repository names, that will be temporared disabled before
        listing packages; this is applied before enabling of repositories

    Worker replies with [pkg1, pkg2, ...].
    """
    __slots__ = ('kind', 'allow_duplicates', 'sort', 'include_repos',
            'exclude_repos')

    SUPPORTED_KINDS = ( 'installed', 'available', 'avail_reinst'
                      , 'avail_notinst', 'all')

    def __init__(self, kind, allow_duplicates, sort=False,
            include_repos=None, exclude_repos=None):
        YumJob.__init__(self)
        if not isinstance(kind, basestring):
            raise TypeError("kind must be a string")
        if not kind in self.SUPPORTED_KINDS:
            raise ValueError("kind must be one of {%s}" %
                    ", ".join(self.SUPPORTED_KINDS))
        for arg in ('include_repos', 'exclude_repos'):
            val = locals()[arg]
            if (   not val is None
               and not isinstance(arg, (tuple, list, basestring))):
                raise TypeError("expected list or string for %s" % arg)
        self.kind = kind
        self.allow_duplicates = bool(allow_duplicates)
        self.sort = bool(sort)
        self.include_repos = include_repos
        self.exclude_repos = exclude_repos

class YumFilterPackages(YumGetPackageList):   #pylint: disable=R0903
    """
    Job similar to YumGetPackageList, but allowing to specify
    filter on packages.
    Arguments (plus those in YumGetPackageList):
      name, epoch, version, release, arch, nevra, envra, evra,
      repoid, exact_match

    Argument ``exact_match`` makes the name property being compared byte
    by byte. If ``False``, all packages containing ``name``'s value either
    in name of summary will match.

    Some of those are redundant, but filtering is optimized for
    speed, so supplying all of them won't affect performance.

    Worker replies with [pkg1, pkg2, ...].
    """
    __slots__ = (
        'name', 'epoch', 'version', 'release', 'arch',
        'nevra', 'envra', 'evra', 'repoid', 'exact_match')

    def __init__(self, kind, allow_duplicates,
            exact_match=True,
            sort=False, include_repos=None, exclude_repos=None,
            name=None, epoch=None, version=None,
            release=None, arch=None,
            nevra=None, evra=None,
            envra=None,
            repoid=None):
        if nevra is not None and not util.RE_NEVRA.match(nevra):
            raise ValueError("Invalid nevra: %s" % nevra)
        if evra is not None and not util.RE_EVRA.match(evra):
            raise ValueError("Invalid evra: %s" % evra)
        if envra is not None and not util.RE_ENVRA.match(evra):
            raise ValueError("Invalid envra: %s" % envra)
        YumGetPackageList.__init__(self, kind, allow_duplicates, sort,
                include_repos=include_repos, exclude_repos=exclude_repos)
        self.name = name
        self.epoch = None if epoch is None else str(epoch)
        self.version = version
        self.release = release
        self.arch = arch
        self.nevra = nevra
        self.evra = evra
        self.envra = envra
        self.repoid = repoid
        self.exact_match = bool(exact_match)

class YumSpecificPackageJob(YumAsyncJob):  #pylint: disable=R0903
    """
    Abstract job taking instance of yumdb.PackageInfo as argument or
    package's nevra.
    Arguments:
      pkg - plays different role depending on job subclass;
            can also be a nevra
    """
    __slots__ = ('pkg', )
    def __init__(self, pkg, async=False, metadata=None):
        if isinstance(pkg, basestring):
            if not util.RE_NEVRA_OPT_EPOCH.match(pkg):
                raise errors.InvalidNevra('not a valid nevra "%s"' % pkg)
        elif not isinstance(pkg, PackageInfo):
            raise TypeError("pkg must be either string or instance"
                " of PackageInfo")
        YumAsyncJob.__init__(self, async=async, metadata=metadata)
        self.pkg = pkg

class YumInstallPackage(YumSpecificPackageJob):   #pylint: disable=R0903
    """
    Job requesting installation of specific package.
    pkg argument should be available.
    Arguments:
      pkg - same as in YumSpecificPackageJob
      force is a boolean saying:
        True  -> reinstall the package if it's already installed
        False -> fail if the package is already installed

    Worker replies with new instance of package.
    """
    __slots__ = ('force', )
    def __init__(self, pkg, async=False, force=False, metadata=None):
        YumSpecificPackageJob.__init__(
                self, pkg, async=async, metadata=metadata)
        self.force = bool(force)

class YumRemovePackage(YumSpecificPackageJob):    #pylint: disable=R0903
    """
    Job requesting removal of specific package.
    pkg argument should be installed.
    """
    pass

class YumUpdateToPackage(YumSpecificPackageJob):  #pylint: disable=R0903
    """
    Job requesting update to provided specific package.
    Package is updated to epoch, version and release of this
    provided available package.

    Worker replies with new instance of package.
    """
    pass

class YumUpdatePackage(YumSpecificPackageJob):    #pylint: disable=R0903
    """
    Job requesting update of package, optionally reducing possible
    candidate packages to ones with specific evr.
    Arguments:
      to_epoch, to_version, to_release
      force is a boolean, that has meaning only when update_only is False:
        True  -> reinstall the package if it's already installed
        False -> fail if the package is already installed

    The arguments more given, the more complete filter of candidates.

    Worker replies with new instance of package.
    """
    __slots__ = ('to_epoch', 'to_version', 'to_release', 'force')

    def __init__(self, pkg, async=False,
            to_epoch=None, to_version=None, to_release=None, force=False,
            metadata=None):
        if not isinstance(pkg, (basestring, PackageInfo)):
            raise TypeError("pkg must be either instance of yumdb.PackageInfo"
                    " or nevra string")
        YumSpecificPackageJob.__init__(
                self, pkg, async=async, metadata=metadata)
        self.to_epoch = to_epoch
        self.to_version = to_version
        self.to_release = to_release
        self.force = bool(force)

class YumCheckPackage(YumSpecificPackageJob):       #pylint: disable=R0903
    """
    Request verification information for instaled package and its files.

    Arguments:
        pkg - either instance of PackageInfo or nevra string.
              In latter case it will be replaced for YumWorker with instance
              of PackageInfo.

    Worker replies with ``(pkg_info, pkg_check)``.
      where:
        ``pkg_info``  - is instance of PackageInfo
        ``pkg_check`` - new instance of yumdb.PackageCheck
    """
    def __init__(self, pkg, async=False, metadata=None):
        YumSpecificPackageJob.__init__(self, pkg, async=async,
                metadata=metadata)
        if isinstance(pkg, PackageInfo) and not pkg.installed:
            raise ValueError("package must be installed to check it")

class YumCheckPackageFile(YumCheckPackage):     #pylint: disable=R0903
    """
    Request verification information for particular file of installed
    package.

    Worker replies with ``(pkg_info, pkg_check)``.
      where:
        ``pkg_info``  - is instance of PackageInfo
        ``pkg_check`` - new instance of yumdb.PackageCheck containing only
                        requested file.
    """
    __slots__ = ('file_name', )
    def __init__(self, pkg, file_name, *args, **kwargs):
        YumCheckPackage.__init__(self, pkg, *args, **kwargs)
        if not isinstance(file_name, basestring):
            raise TypeError("file_name must be string")
        self.file_name = file_name

class YumInstallPackageFromURI(YumAsyncJob):     #pylint: disable=R0903
    """
    Job requesting installation of specific package from URI.
    Arguments:
      uri is either a path to rpm package on local filesystem or url
        of rpm stored on remote host
      update_only is a boolean:
        True  -> install the package only if the older version is installed
        False -> install the package if it's not already installed
      force is a boolean, that has meaning only when update_only is False:
        True  -> reinstall the package if it's already installed
        False -> fail if the package is already installed

    Worker replies with new instance of package.
    """
    __slots__ = ('uri', 'update_only', "force")
    def __init__(self, uri, async=False, update_only=False, force=False,
            metadata=None):
        if not isinstance(uri, basestring):
            raise TypeError("uri must be a string")
        if uri.startswith('file://'):
            uri = uri[len('file://'):]
        if not yum.misc.re_remote_url(uri) and not os.path.exists(uri):
            raise errors.InvalidURI(uri)
        YumAsyncJob.__init__(self, async=async, metadata=metadata)
        self.uri = uri
        self.update_only = bool(update_only)
        self.force = bool(force)

class YumGetRepositoryList(YumJob):                 #pylint: disable=R0903
    """
    Job requesing a list of repositories.
    Arguments:
      kind - supported values are in SUPPORTED_KINDS tuple

    Worker replies with [repo1, repo2, ...].
    """
    __slots__ = ('kind', )

    SUPPORTED_KINDS = ('all', 'enabled', 'disabled')

    def __init__(self, kind):
        YumJob.__init__(self)
        if not isinstance(kind, basestring):
            raise TypeError("kind must be a string")
        if not kind in self.SUPPORTED_KINDS:
            raise ValueError("kind must be one of {%s}" %
                    ", ".join(self.SUPPORTED_KINDS))
        self.kind = kind

class YumFilterRepositories(YumGetRepositoryList):  #pylint: disable=R0903
    """
    Job similar to YumGetRepositoryList, but allowing to specify
    filter on packages.
    Arguments (plus those in YumGetRepositoryList):
      name, gpg_check, repo_gpg_check

    Some of those are redundant, but filtering is optimized for
    speed, so supplying all of them won't affect performance.

    Worker replies with [repo1, repo2, ...].
    """
    __slots__ = ('repoid', 'gpg_check', 'repo_gpg_check')

    def __init__(self, kind,
            repoid=None, gpg_check=None, repo_gpg_check=None):
        YumGetRepositoryList.__init__(self, kind)
        self.repoid = repoid
        self.gpg_check = None if gpg_check is None else bool(gpg_check)
        self.repo_gpg_check = (
                None if repo_gpg_check is None else bool(repo_gpg_check))

class YumSpecificRepositoryJob(YumJob):             #pylint: disable=R0903
    """
    Abstract job taking instance of yumdb.Repository as argument.
    Arguments:
      repo - (``Repository`` or ``str``) plays different role depending
             on job subclass
    """
    __slots__ = ('repo', )
    def __init__(self, repo):
        if not isinstance(repo, (Repository, basestring)):
            raise TypeError("repoid must be either instance of"
                " yumdb.Repository or string")
        YumJob.__init__(self)
        self.repo = repo

class YumSetRepositoryEnabled(YumSpecificRepositoryJob):#pylint: disable=R0903
    """
    Job allowing to enable or disable repository.
    Arguments:
      enable - (``boolean``) representing next state
    """
    __slots__ = ('enable', )
    def __init__(self, repo, enable):
        YumSpecificRepositoryJob.__init__(self, repo)
        self.enable = bool(enable)

def log_reply_error(job, reply):
    """
    Raises an exception in case of error occured in worker process
    while processing job.
    """
    if isinstance(reply, (int, long)):
        # asynchronous job
        return
    logger = logging.getLogger(__name__)
    if not isinstance(reply, YumJob):
        raise TypeError('expected instance of YumJob for reply, not "%s"'
                % reply.__class__.__name__)
    if reply.result == YumJob.RESULT_ERROR:
        logger.error("%s failed with error %s: %s",
                job, reply.result_data[0].__name__, str(reply.result_data[1]))
        logger.trace_warn("%s exception traceback:\n%s%s: %s",
                job, "".join(reply.result_data[2]),
                reply.result_data[0].__name__, str(reply.result_data[1]))
        reply.result_data[1].tb_printed = True
        raise reply.result_data[1]
    elif reply.result == YumJob.RESULT_TERMINATED:
        logger.warn('%s terminated', job)
    else:
        logger.debug('%s completed with success', job)




#!/usr/bin/python
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Authors: Jan Safranek <jsafrane@redhat.com>
#

import os;
import pywbem;
import optparse
import re
import cgi
import sys

class HtmlExporter(object):
    
    def __init__(self, cliconn):
        self.classcache = {}
        self.cliconn = cliconn
        # original list of classes
        self.classes = []
        # set of all classes to print, i.e. incl. all parents
        self.queue = set()
        self.file = None

    def escape(self, string):
        re_string = re.compile(r'(?P<htmlchars>[<&>])|(?P<space>^[ \t]+)|(?P<lineend>\r\n|\r|\n)|(?P<protocal>(^|\s)((http|ftp)://.*?))(\s|$)', re.S|re.M|re.I)
        re_nl = re.compile(r"\\n")
        def do_sub(m):
            c = m.groupdict()
            if c['htmlchars']:
                return cgi.escape(c['htmlchars'])
            if c['lineend']:
                return '<br>'
            elif c['space']:
                t = m.group().replace('\t', '&nbsp;'*tabstop)
                t = t.replace(' ', '&nbsp;')
                return t
            elif c['space'] == '\t':
                return ' '*tabstop;
            else:
                url = m.group('protocal')
                if url.startswith(' '):
                    prefix = ' '
                    url = url[1:]
                else:
                    prefix = ''
                last = m.groups()[-1]
                if last in ['\n', '\r', '\r\n']:
                    last = '<br>'
                return '%s<a href="%s">%s</a>%s' % (prefix, url, url, last)
        string2 = re.sub(re_string, do_sub, string)
        return re.sub(re_nl, "<br/>", string2)
    
    def load_class(self, classname):
        if classname in self.classcache:
            return self.classcache[classname]

        c = self.cliconn.GetClass(classname)
        self.classcache[classname] = c
        return c

    def source_class(self, c, param):
        """
            Find a class where a CIMMethod or CIMProperty is first defined.
            Start at class c and inspect all parents.
        """

        while True:
            if not c.superclass:
                # we're at the top class
                return c.classname

            if isinstance(param, pywbem.CIMMethod):
                if not c.methods.has_key(param.name):
                    break
            else:
                if not c.properties.has_key(param.name):
                    break
            parent = c
            c = self.load_class(c.superclass)
        return parent.classname

    def display_type(self, param):
        """
            Return displayable type of given parameter.
            It adds [] if it's array and class name of referenced classes.
        """
        url=None
        if param.reference_class:
            ptype = param.reference_class
            url=param.reference_class + ".html"
        else:
            ptype = param.type
        if param.is_array:
            ptype = ptype + "[]"
        if url:
            ptype = "<a href=\"%s\">%s</a>" % (url, ptype)
        return ptype

    def print_parameters(self, params):
        """
            Print table of method parametes.
        """
        if not params:
            print >>self.file, "None"
            return

        print >>self.file, "<table class=\"parameters\">"
        for p in params.values():
            direction = set()
            if p.qualifiers.has_key("Out"):
                direction.add("OUT")
            if p.qualifiers.has_key("In"):
                if p.qualifiers["In"].value:
                    direction.add("IN")
            if not direction:
                direction.add("IN")
            direction = "/".join(sorted(direction))

            print >>self.file, "<tr>"
            print >>self.file, "<td>%s</td><td>%s</td><td><b>%s</b></td>" % (direction, self.display_type(p), p.name)
            print >>self.file, "<td><table class=\"qualifiers\">"
            self.print_qualifiers(p.qualifiers.values())
            print >>self.file, "</table></td>"
            print >>self.file, "</tr>"
        print >>self.file, "</table>"

    def print_qualifiers(self, qualifiers):
        """
            Print contenf of table of qualifiers.
            Only Deprecated, Description, Values and ValueMap are recognized.
        """
        deprecated = ""
        for q in sorted(qualifiers):
            if q.name == "Deprecated":
                deprecated = "<div class=\"deprecated\">DEPRECATED</div> "
            if q.name == "Description":
                print >>self.file, "<tr><td class=\"qualifiers\" colspan=\"2\">%s%s</td></tr>" % (deprecated, self.escape(q.value))
            if q.name == "Values":
                print >>self.file, "<tr><td class=\"qualifiers\"><b>Values</b></td> <td class=\"qualifiers\"><table><tr><td class=\"qualifiers\">%s</td></tr></table></td></tr>" % ("</td></tr><tr><td class=\"qualifiers\">".join(q.value))
            if q.name == "ValueMap":
                print >>self.file, "<tr><td class=\"qualifiers\"><b>ValueMap</b></td> <td class=\"qualifiers\"><table><tr><td class=\"qualifiers\">%s</td></tr></table></td></tr>" % ("</td></tr><tr><td class=\"qualifiers\">".join(q.value))

    def compare_properties(self, p1, p2):
        """
            Compare two properties if they should printed in Inherited properties.
            Only Name, Description and Implemented are checked.
            Returns False, if the property should be printed in Local.
        """
        if p1.name != p2.name:
            return False
        d1 = p1.qualifiers.get("Description", None)
        d2 = p2.qualifiers.get("Description", None)
        if d1.value != d2.value:
            return False
        i1 = p1.qualifiers.get("Implemented", None)
        i2 = p2.qualifiers.get("Implemented", None)
        if i1 and i1.value and not (i2 and i2.value):
            return False
        return True

    def print_keys(self, c):
        print >>self.file, "<h3>Key properties:</h3>"
        print >>self.file, "<table class=\"key\">"
        for prop in c.properties.values():
            if prop.qualifiers.has_key('Key'):
                src = self.source_class(c, prop)
                print >>self.file, "<tr class=\"key\"><td class=\"key\"><a href=\"%s.html#%s\">%s</a></td></tr>" % (src, prop.name, prop.name)
        print >>self.file, "</table>"

    def print_class(self, c):
        """
            Print one class, inc. header.
        """
        parent = None
        if c.superclass:
            parent = self.load_class(c.superclass)
        print >>self.file, "<h2>%s : <a href=\"%s.html\">%s</a></h2>" % (c.classname, c.superclass, c.superclass)

        description = c.qualifiers.get("Description", None)
        if not description:
            description = parent.qualifiers.get("Description", None)
        if description:
            print >>self.file, "<p>%s</p>" % (self.escape(description.value))

        self.print_keys(c)

        local_props = []
        inherited_props = []
        for name in sorted(c.properties.keys()):
            if parent and parent.properties.has_key(name):
                inherited_props.append(c.properties[name])
                if not self.compare_properties(c.properties[name], parent.properties[name]):
                    # the property was overridden
                    local_props.append(c.properties[name])
            else:
                local_props.append(c.properties[name])

        local_methods = []
        inherited_methods = []
        for name in sorted(c.methods.keys()):
            if parent and parent.methods.has_key(name):
                inherited_methods.append(c.methods[name])
                if not self.compare_properties(c.methods[name], parent.methods[name]):
                    # the property was overridden
                    local_methods.append(c.methods[name])
            else:
                local_methods.append(c.methods[name])

        print >>self.file, "<h3>Local properties:</h3>"

        if local_props:
            print >>self.file, "<table class=\"properties\"> <tr> <th>Name</th> <th>Type</th> <th>Qualifiers</th>"

            for prop in local_props:
                if prop in inherited_props:
                    style="i"
                else:
                    style="b"
                print >>self.file, "<tr> <td><%s><a name=\"%s\"/>%s</%s></td> <td>%s</td><td>" % (style, prop.name, prop.name, style, self.display_type(prop))
                print >>self.file, "<table class=\"qualifiers\">"
                self.print_qualifiers(prop.qualifiers.values())
                print >>self.file, "</table>"
                print >>self.file, "</td></tr>"
            print >>self.file, "</table>"
            print >>self.file, "<small><b>bold</b> - new property, <i>italics</i> - overridden property, usually with new description.</small>"
        else:
            print >>self.file, "None"

        print >>self.file, "<h3>Local methods:</h3>"
        if local_methods:
            print >>self.file, "<table class=\"properties\"> <tr> <th>Type</th><th>Name</th><th>Qualifiers</th>"
            for m in local_methods:
                print >>self.file, "<tr> <td>%s</td> <td><b><a name=\"%s\"/>%s</b></td><td>" % (m.return_type, m.name, m.name)
                print >>self.file, "<table class=\"qualifiers\">"
                self.print_qualifiers(m.qualifiers.values())
                print >>self.file, "<tr><td class=\"qualifiers\"><b>Parameters</b></td><td class=\"qualifiers\">"
                self.print_parameters(m.parameters)
                print >>self.file, "</td></tr>"
                print >>self.file, "</table>"
                print >>self.file, "</td></tr>"
            print >>self.file, "</table>"
        else:
            print >>self.file, "None"

        print >>self.file, "<h3>Inherited properties:</h3>"
        if inherited_props:
            print >>self.file, "<table class=\"inherited\"> <tr> <th>Name</th><th>Type</th><th>Source</th>"
            for p in inherited_props:
                src = self.source_class(c, p)
                print >>self.file, "<tr><td><a href=\"%s.html#%s\">%s</a></td><td>%s</td><td>%s</td></tr>" % (src, p.name, p.name, self.display_type(p), src)
            print >>self.file, "</table>"
        else:
            print >>self.file, "None"

        print >>self.file, "<h3>Inherited methods:</h3>"
        if inherited_methods:
            print >>self.file, "<table class=\"inherited\"> <tr> <th>Name</th><th>Source</th>"
            for m in inherited_methods:
                src = self.source_class(c, m)
                print >>self.file, "<tr><td><a href=\"%s.html#%s\">%s</a></td><td>%s</td></tr>" % (src, m.name, m.name, src)
            print >>self.file, "</table>"
        else:
            print >>self.file, "None"



    def print_file(self, filename):
        f = open(filename, "r")
        data = f.read()
        self.file.write(data)
        f.close()

    def print_page(self, c, header, footer):
        print "exporting ", c.classname
        self.file = open(c.classname + ".html", "w")
        print >>self.file, "<html><head><title>%s</title>" % (c.classname)
        print >>self.file, "<link rel=\"stylesheet\" type=\"text/css\" charset=\"utf-8\" media=\"all\" href=\"mof.css\" />"
        print >>self.file, "</head><body>"
        if header:
            self.print_file(header)
        self.print_class(c)
        if footer:
            self.print_file(footer)
        print >>self.file, "</body></html>"
        self.file.close()
        
    def add_class(self, classname):
        self.classes.append(classname)
        print "adding", classname
        self.queue.add(classname)
        # add all parents
        c = self.load_class(classname)
        parentname = c.superclass
        while parentname:
            if parentname in self.queue:
                break
            print "adding", parentname
            self.queue.add(parentname)
            c = self.load_class(parentname)
            parentname = c.superclass

    def export(self, header=None, footer=None):
        """
            Print everything, i.e. index.html + all classes + all their parents.
        """
        # remove duplicate classes from the queue (and sort them)
        self.queue = sorted(self.queue)
        self.print_index(header, footer)
        while self.queue:
            c = self.queue.pop()
            self.print_page(self.load_class(c), header, footer)

    def _do_print_tree(self, name, subtree, level):
        print >>self.file, "<tr class=\"tree\"><td class=\"tree\">"
        if level:
            print >>self.file, "&nbsp;&nbsp;&nbsp;|"*level + "--- "
        if name in self.classes:
            printname = "<span class=\"tree_local\">%s</span>" % (name)
        else:
            printname = "<span class=\"tree_inherited\">%s</span>" % (name)
        print >>self.file, "<a href=\"%s.html\">%s</a>" % (name, printname)
        print >>self.file, "</td></tr>"
        for n in subtree.keys():
            self._do_print_tree(n, subtree[n], level+1)

    def print_tree(self):
        """
            Create inheritance tree of classes.
        """
        # hash classname -> list of hash of (direct) sublasses 
        subclasses = {}
        # hash classname -> nr. of its parents
        parents = {}
        # initialize the hash
        for cname in self.queue:
            subclasses[cname] = {}
            parents[cname] = 0

        # fill the hash
        for cname in self.queue:
            c = self.load_class(cname)
            if c.superclass:
                subclasses[c.superclass][cname] = subclasses[cname]
                parents[cname] += 1

        print >>self.file, "<table class=\"tree\">"
        # print all top classes
        for cname in self.queue:
            if parents[cname] == 0:
                self._do_print_tree(cname, subclasses[cname], 0)
        print >>self.file, "</table>"

    def print_index(self, header, footer):
        print "exporting index"
        self.file = open("index.html", "w")
        print >>self.file, "<html><head><title>Index</title>"
        print >>self.file, "<link rel=\"stylesheet\" type=\"text/css\" charset=\"utf-8\" media=\"all\" href=\"mof.css\" />"
        print >>self.file, "</head><body>"
        if header:
            self.print_file(header)
        self.print_tree()
        if footer:
            self.print_file(footer)
        print >>self.file, "</body></html>"
        self.file.close()
            
description = """
Generate HTML documentation for given CIM classes. The tool connects to specified CIMOM
and reads class definition from there. It generates separate HTML file for each class
specified on command line and for all it's parents.

The tool also generates index.html with inheritance tree.

All generated HTML pages will optionally contain header and/or footer. Both must be valid
HTML snippets and will be inserted just after <body> or before </body>.
"""

parser = optparse.OptionParser(usage="usage: %prog [options] classname [classname ...]", description=description)
parser.add_option('-u', '--url', action='store', dest='addr', default='https://localhost:5989', help='URL of CIM server, default: https://localhost:5989')
parser.add_option('-U', '--user', action='store', dest='user', default=None, help='CIM user name')
parser.add_option('-P', '--password', action='store', dest='password', default=None, help='CIM password')
parser.add_option('-H', '--header', action='store', dest='header', default=None, help='File with HTML page header')
parser.add_option('-F', '--footer', action='store', dest='footer', default=None, help='File with HTML page footer')

(options, args) = parser.parse_args()

cliconn = pywbem.WBEMConnection(options.addr, (options.user, options.password))
exporter = HtmlExporter(cliconn)
for c in args:
    exporter.add_class(c)
exporter.export(options.header, options.footer)




# -*- encoding: utf-8 -*-
# Software Management Providers
#
# Copyright (C) 2012-2013 Red Hat, Inc.  All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

"""
Just a common functionality related to SoftwareInstallationJob
and SoftwareVerificationJob.
"""

from datetime import datetime, timedelta
import pywbem
import time

from lmi.providers import cmpi_logging
from lmi.software import util
from lmi.software.core import Error
from lmi.software.yumdb import errors
from lmi.software.yumdb import jobs
from lmi.software.yumdb import YumDB

JOB_CLASS_NAMES = (
    "LMI_SoftwareInstallationJob",
    "LMI_SoftwareVerificationJob"
)

JOB_DESCRIPTIONS = {
    jobs.YumInstallPackage :
        'Software package installation job %(jobid)d for "%(pkg)s".',
    jobs.YumRemovePackage :
        'Software package removal job %(jobid)d for "%(pkg)s".',
    jobs.YumUpdateToPackage :
        'Software package update job %(jobid)d for "%(pkg)s".',
    jobs.YumUpdatePackage :
        'Software package update job %(jobid)d for "%(pkg)s".',
    jobs.YumInstallPackageFromURI :
        'Software package installation job %(jobid)d from uri: "%(uri)s".',
    jobs.YumCheckPackage :
        'Software package check job %(jobid)d for "%(pkg)s".',
    jobs.YumCheckPackageFile :
        'File verification job %(jobid)d for package "%(pkg)s".',
}

# identificators of InstallationService method, which may trigger
# creation of asynchronous job
( JOB_METHOD_INSTALL_FROM_SOFTWARE_IDENTITY
, JOB_METHOD_INSTALL_FROM_URI
, JOB_METHOD_INSTALL_FROM_BYTE_STREAM
, JOB_METHOD_VERIFY_INSTALLED_IDENTITY) = range(4)

# above identificators point to this array to their description
JOB_METHOD_NAMES = (
        "InstallFromSoftwareIdentity",
        "InstallFromURI",
        "InstallFromByteStream",
        "VerifyInstalledIdentity")

LOG = cmpi_logging.get_logger(__name__)

class Values(object):
    class DetailedStatus(object):
        Not_Available = pywbem.Uint16(0)
        No_Additional_Information = pywbem.Uint16(1)
        Stressed = pywbem.Uint16(2)
        Predictive_Failure = pywbem.Uint16(3)
        Non_Recoverable_Error = pywbem.Uint16(4)
        Supporting_Entity_in_Error = pywbem.Uint16(5)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Not Available',
                1: 'No Additional Information',
                2: 'Stressed',
                3: 'Predictive Failure',
                4: 'Non-Recoverable Error',
                5: 'Supporting Entity in Error'
        }

    class Status(object):
        OK = 'OK'
        Error = 'Error'
        Degraded = 'Degraded'
        Unknown = 'Unknown'
        Pred_Fail = 'Pred Fail'
        Starting = 'Starting'
        Stopping = 'Stopping'
        Service = 'Service'
        Stressed = 'Stressed'
        NonRecover = 'NonRecover'
        No_Contact = 'No Contact'
        Lost_Comm = 'Lost Comm'
        Stopped = 'Stopped'

    class HealthState(object):
        Unknown = pywbem.Uint16(0)
        OK = pywbem.Uint16(5)
        Degraded_Warning = pywbem.Uint16(10)
        Minor_failure = pywbem.Uint16(15)
        Major_failure = pywbem.Uint16(20)
        Critical_failure = pywbem.Uint16(25)
        Non_recoverable_error = pywbem.Uint16(30)
        # DMTF_Reserved = ..
        # Vendor_Specific = 32768..65535
        _reverse_map = {
                0: 'Unknown',
                5: 'OK',
                10: 'Degraded/Warning',
                15: 'Minor failure',
                20: 'Major failure',
                25: 'Critical failure',
                30: 'Non-recoverable error'
        }

    class JobState(object):
        New = pywbem.Uint16(2)
        Starting = pywbem.Uint16(3)
        Running = pywbem.Uint16(4)
        Suspended = pywbem.Uint16(5)
        Shutting_Down = pywbem.Uint16(6)
        Completed = pywbem.Uint16(7)
        Terminated = pywbem.Uint16(8)
        Killed = pywbem.Uint16(9)
        Exception = pywbem.Uint16(10)
        Service = pywbem.Uint16(11)
        Query_Pending = pywbem.Uint16(12)
        # DMTF_Reserved = 13..32767
        # Vendor_Reserved = 32768..65535
        _reverse_map = {
                2: 'New',
                3: 'Starting',
                4: 'Running',
                5: 'Suspended',
                6: 'Shutting Down',
                7: 'Completed',
                8: 'Terminated',
                9: 'Killed',
                10: 'Exception',
                11: 'Service',
                12: 'Query Pending'
        }

    class GetError(object):
        Success = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unspecified_Error = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Access_Denied = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        # Vendor_Specific = 32768..65535

    class KillJob(object):
        Success = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unknown = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Access_Denied = pywbem.Uint32(6)
        Not_Found = pywbem.Uint32(7)
        # DMTF_Reserved = ..
        # Vendor_Specific = 32768..65535

    class RecoveryAction(object):
        Unknown = pywbem.Uint16(0)
        Other = pywbem.Uint16(1)
        Do_Not_Continue = pywbem.Uint16(2)
        Continue_With_Next_Job = pywbem.Uint16(3)
        Re_run_Job = pywbem.Uint16(4)
        Run_Recovery_Job = pywbem.Uint16(5)
        _reverse_map = {
                0: 'Unknown',
                1: 'Other',
                2: 'Do Not Continue',
                3: 'Continue With Next Job',
                4: 'Re-run Job',
                5: 'Run Recovery Job'
        }

    class RunDayOfWeek(object):
        _Saturday = pywbem.Sint8(-7)
        _Friday = pywbem.Sint8(-6)
        _Thursday = pywbem.Sint8(-5)
        _Wednesday = pywbem.Sint8(-4)
        _Tuesday = pywbem.Sint8(-3)
        _Monday = pywbem.Sint8(-2)
        _Sunday = pywbem.Sint8(-1)
        ExactDayOfMonth = pywbem.Sint8(0)
        Sunday = pywbem.Sint8(1)
        Monday = pywbem.Sint8(2)
        Tuesday = pywbem.Sint8(3)
        Wednesday = pywbem.Sint8(4)
        Thursday = pywbem.Sint8(5)
        Friday = pywbem.Sint8(6)
        Saturday = pywbem.Sint8(7)
        _reverse_map = {
                0: 'ExactDayOfMonth',
                1: 'Sunday',
                2: 'Monday',
                3: 'Tuesday',
                4: 'Wednesday',
                5: 'Thursday',
                6: 'Friday',
                7: 'Saturday',
                -1: '-Sunday',
                -7: '-Saturday',
                -6: '-Friday',
                -5: '-Thursday',
                -4: '-Wednesday',
                -3: '-Tuesday',
                -2: '-Monday'
        }

    class RunMonth(object):
        January = pywbem.Uint8(0)
        February = pywbem.Uint8(1)
        March = pywbem.Uint8(2)
        April = pywbem.Uint8(3)
        May = pywbem.Uint8(4)
        June = pywbem.Uint8(5)
        July = pywbem.Uint8(6)
        August = pywbem.Uint8(7)
        September = pywbem.Uint8(8)
        October = pywbem.Uint8(9)
        November = pywbem.Uint8(10)
        December = pywbem.Uint8(11)
        _reverse_map = {
                0: 'January',
                1: 'February',
                2: 'March',
                3: 'April',
                4: 'May',
                5: 'June',
                6: 'July',
                7: 'August',
                8: 'September',
                9: 'October',
                10: 'November',
                11: 'December'
        }

    class GetErrors(object):
        Success = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unspecified_Error = pywbem.Uint32(2)
        Timeout = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        Access_Denied = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        # Vendor_Specific = 32768..65535

    class CommunicationStatus(object):
        Unknown = pywbem.Uint16(0)
        Not_Available = pywbem.Uint16(1)
        Communication_OK = pywbem.Uint16(2)
        Lost_Communication = pywbem.Uint16(3)
        No_Contact = pywbem.Uint16(4)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'Not Available',
                2: 'Communication OK',
                3: 'Lost Communication',
                4: 'No Contact'
        }

    class OperationalStatus(object):
        Unknown = pywbem.Uint16(0)
        Other = pywbem.Uint16(1)
        OK = pywbem.Uint16(2)
        Degraded = pywbem.Uint16(3)
        Stressed = pywbem.Uint16(4)
        Predictive_Failure = pywbem.Uint16(5)
        Error = pywbem.Uint16(6)
        Non_Recoverable_Error = pywbem.Uint16(7)
        Starting = pywbem.Uint16(8)
        Stopping = pywbem.Uint16(9)
        Stopped = pywbem.Uint16(10)
        In_Service = pywbem.Uint16(11)
        No_Contact = pywbem.Uint16(12)
        Lost_Communication = pywbem.Uint16(13)
        Aborted = pywbem.Uint16(14)
        Dormant = pywbem.Uint16(15)
        Supporting_Entity_in_Error = pywbem.Uint16(16)
        Completed = pywbem.Uint16(17)
        Power_Mode = pywbem.Uint16(18)
        Relocating = pywbem.Uint16(19)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'Other',
                2: 'OK',
                3: 'Degraded',
                4: 'Stressed',
                5: 'Predictive Failure',
                6: 'Error',
                7: 'Non-Recoverable Error',
                8: 'Starting',
                9: 'Stopping',
                10: 'Stopped',
                11: 'In Service',
                12: 'No Contact',
                13: 'Lost Communication',
                14: 'Aborted',
                15: 'Dormant',
                16: 'Supporting Entity in Error',
                17: 'Completed',
                18: 'Power Mode',
                19: 'Relocating'
        }

    class OperatingStatus(object):
        Unknown = pywbem.Uint16(0)
        Not_Available = pywbem.Uint16(1)
        Servicing = pywbem.Uint16(2)
        Starting = pywbem.Uint16(3)
        Stopping = pywbem.Uint16(4)
        Stopped = pywbem.Uint16(5)
        Aborted = pywbem.Uint16(6)
        Dormant = pywbem.Uint16(7)
        Completed = pywbem.Uint16(8)
        Migrating = pywbem.Uint16(9)
        Emigrating = pywbem.Uint16(10)
        Immigrating = pywbem.Uint16(11)
        Snapshotting = pywbem.Uint16(12)
        Shutting_Down = pywbem.Uint16(13)
        In_Test = pywbem.Uint16(14)
        Transitioning = pywbem.Uint16(15)
        In_Service = pywbem.Uint16(16)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'Not Available',
                2: 'Servicing',
                3: 'Starting',
                4: 'Stopping',
                5: 'Stopped',
                6: 'Aborted',
                7: 'Dormant',
                8: 'Completed',
                9: 'Migrating',
                10: 'Emigrating',
                11: 'Immigrating',
                12: 'Snapshotting',
                13: 'Shutting Down',
                14: 'In Test',
                15: 'Transitioning',
                16: 'In Service'
        }

    class LocalOrUtcTime(object):
        Local_Time = pywbem.Uint16(1)
        UTC_Time = pywbem.Uint16(2)
        _reverse_map = {
                1: 'Local Time',
                2: 'UTC Time'
        }

    class RequestStateChange(object):
        Completed_with_No_Error = pywbem.Uint32(0)
        Not_Supported = pywbem.Uint32(1)
        Unknown_Unspecified_Error = pywbem.Uint32(2)
        Can_NOT_complete_within_Timeout_Period = pywbem.Uint32(3)
        Failed = pywbem.Uint32(4)
        Invalid_Parameter = pywbem.Uint32(5)
        In_Use = pywbem.Uint32(6)
        # DMTF_Reserved = ..
        Method_Parameters_Checked___Transition_Started = pywbem.Uint32(4096)
        Invalid_State_Transition = pywbem.Uint32(4097)
        Use_of_Timeout_Parameter_Not_Supported = pywbem.Uint32(4098)
        Busy = pywbem.Uint32(4099)
        # Method_Reserved = 4100..32767
        # Vendor_Specific = 32768..65535
        class RequestedState(object):
            Start = pywbem.Uint16(2)
            Suspend = pywbem.Uint16(3)
            Terminate = pywbem.Uint16(4)
            Kill = pywbem.Uint16(5)
            Service = pywbem.Uint16(6)
            # DMTF_Reserved = 7..32767
            # Vendor_Reserved = 32768..65535

    class PrimaryStatus(object):
        Unknown = pywbem.Uint16(0)
        OK = pywbem.Uint16(1)
        Degraded = pywbem.Uint16(2)
        Error = pywbem.Uint16(3)
        # DMTF_Reserved = ..
        # Vendor_Reserved = 0x8000..
        _reverse_map = {
                0: 'Unknown',
                1: 'OK',
                2: 'Degraded',
                3: 'Error'
        }

@cmpi_logging.trace_function
def get_verification_out_params(job):
    """
    Get the output parameters for verification job. They may not be computed
    yet. In that case compute them a update the job in YumWorker process.

    :param job: (``jobs.YumCheckPackage``)
    :rtype: (``dict``) Dictionary of output parameters with pywbem values.
    """
    if not isinstance(job, jobs.YumCheckPackage):
        raise TypeError("job must be a YumCheckPackage instance")
    if (   not job.metadata.get("output_params", [])
       and job.state == job.COMPLETED):
        from lmi.software.core import IdentityFileCheck
        failed = []
        pkg_info, pkg_check = job.result_data
        for file_name in pkg_check:
            pkg_file = pkg_check[file_name]
            file_check = IdentityFileCheck.test_file(pkg_info,
                    pkg_check.file_checksum_type, pkg_file)
            if not IdentityFileCheck.file_check_passed(file_check):
                failed.append(IdentityFileCheck.file_check2model(
                    file_check, job=job))
        metadata = {
            'output_params' : {
                'Failed' : pywbem.CIMProperty("Failed",
                    type="reference", is_array=True, value=failed)
            }
        }
        # update local instance
        job.update(metadata=metadata)
    return job.metadata.get('output_params', [])

@cmpi_logging.trace_function
def make_method_params(job, class_name, include_input, include_output):
    """
    Create a class of given name with all input or output parameters
    of the asynchronous method. Typically used to assemble
    CIM_ConcreteJob.JobInParameters or CIM_InstMethodCall.MethodParameters
    values.

    :param job: (``YumJob``) Instance of job created as a result of method
        invocation. It carries method parameters.
    :param class_name: (``str``) Name of the class to create.
    :param include_input: (``bool``) Whether input parameters should be
        included in the returned class.
    :param include_output: (``bool``) Whether output parameters should be
        included in the returned class.
    :rtype: CIMInstance of the created class.
    """
    path = util.new_instance_name(class_name)
    inst = pywbem.CIMInstance(classname=class_name, path=path)
    if include_input and "input_params" in job.metadata:
        for (name, value) in job.metadata["input_params"].items():
            inst[name] = value
    if include_output:
        if isinstance(job, jobs.YumCheckPackage):
            # make sure, that output parameters are computed
            # TODO: uncomment this, when pegasus properly handles instances
            # of unknown classes - we can not create Failed property, which
            # is an array of references in association class representing
            # result of asynchronous method
            #get_verification_out_params(job)
            pass
        if "output_params" in job.metadata:
            # overwrite any input parameter
            for (name, value) in job.metadata["output_params"].iteritems():
                inst[name] = value
        return_value = make_return_value(job)
        if return_value is not None:
            inst["__ReturnValue"] = return_value
    return inst

def make_return_value(job):
    """
    Compute return value of particular job.

    :param job: (``jobs.YumAsyncJob``) Asynchronous job.
    :rtype: (``pywbem.Uint32``) Return value of asynchronous method or ``None``
        if job has not completed yet.
    """
    from lmi.software.core import InstallationService

    if not isinstance(job, jobs.YumAsyncJob):
        raise TypeError("job must be a YumAsyncJob")

    if job.state == job.COMPLETED:
        return InstallationService. \
                Values.InstallFromURI.Job_Completed_with_No_Error
    if job.state == job.EXCEPTION:
        if issubclass(job.result_data[0], (
                errors.InvalidNevra, errors.InvalidURI,
                errors.PackageNotFound)):
            return InstallationService.Values.InstallFromURI.Invalid_Parameter
        else:
            return InstallationService.Values.InstallFromURI.Failed
    if job.state == job.TERMINATED:
        return InstallationService.Values.InstallFromURI.Unspecified_Error

    # job has not finished yet
    return None

def job_class2cim_class_name(jobcls):
    """
    Here we map classes of job objects to their corresponding CIM class
    name.

    :param jobcls: (``type``) Subclass of jobs.YumJob.
    """
    if not issubclass(jobcls, (
            jobs.YumInstallPackageFromURI,
            jobs.YumSpecificPackageJob)):
        raise ValueError("Job class \"%s\" does not have any associated"
            " CIM class." % jobcls.__name__)
    if issubclass(jobcls, (jobs.YumCheckPackage, jobs.YumCheckPackageFile)):
        return "LMI_SoftwareVerificationJob"
    return "LMI_SoftwareInstallationJob"

@cmpi_logging.trace_function
def _fill_nonkeys(job, model):
    """
    Fills into the model of instance all non-key properties.
    """
    model['Caption'] = 'Software installation job with id=%d' % job.jobid
    model['CommunicationStatus'] = Values.CommunicationStatus.Not_Available
    model['DeleteOnCompletion'] = job.delete_on_completion
    try:
        description = JOB_DESCRIPTIONS[job.__class__]
        kwargs = job.job_kwargs
        kwargs['jobid'] = job.jobid
        model['Description'] = description % kwargs
    except KeyError:
        LOG().error('no description string found for job class %s',
                job.__class__.__name__)
        model['Description'] = pywbem.CIMProperty('Description',
                type='string', value=None)
    if job.started:
        if job.finished:
            elapsed = job.finished - job.started
        else:
            elapsed = time.time() - job.started
        model['ElapsedTime'] = pywbem.CIMDateTime(timedelta(seconds=elapsed))
    else:
        model["ElapsedTime"] = pywbem.CIMProperty('ElapsedTime',
                type='datetime', value=None)
    model['ErrorCode'] = pywbem.Uint16(0 if job.state != job.EXCEPTION else 1)
    try:
        model['JobState'], model['OperationalStatus'], model['JobStatus'] = {
            jobs.YumJob.NEW        : (Values.JobState.New,
                [Values.OperationalStatus.Dormant], 'Enqueued'),
            jobs.YumJob.RUNNING    : (Values.JobState.Running,
                [Values.OperationalStatus.OK], 'Running'),
            jobs.YumJob.TERMINATED : (Values.JobState.Terminated,
                [Values.OperationalStatus.Stopped], 'Terminated'),
            jobs.YumJob.EXCEPTION  : (Values.JobState.Exception
                , [Values.OperationalStatus.Error]
                , 'Failed'),
            jobs.YumJob.COMPLETED  : (Values.JobState.Completed
                , [ Values.OperationalStatus.OK
                  , Values.OperationalStatus.Completed]
                , 'Finished successfully')
            }[job.state]
    except KeyError:
        LOG().error('unknown job state: %s', job.state)
        model['JobState'] = pywbem.CIMProperty('JobState',
                type='uint16', value=None)
        model['OperationalStatus'] = [Values.OperationalStatus.Unknown]
        model['JobStatus'] = 'Unknown'
    # TODO: uncomment this, when Pegasus propertly supports instances of
    # unknown classes
    #model["JobInParameters"] = make_method_params(
    #    job, "__JobInParameters", True, False)
    method_name = JOB_METHOD_NAMES[job.metadata["method"]]
    model["JobOutParameters"] = make_method_params(
        job, "__MethodParameters_"+method_name+"_Result", False, True)
    if 'method' in job.metadata:
        model['MethodName'] = method_name
    else:
        model["MethodName"] = pywbem.CIMProperty('MethodName',
                type='string', value=None)
    model['Name'] = job.metadata['name']
    model['LocalOrUtcTime'] = Values.LocalOrUtcTime.UTC_Time
    model['PercentComplete'] = pywbem.Uint16(
            100 if job.state == job.COMPLETED else (
            50  if job.state == job.RUNNING   else
            0))
    model['Priority'] = pywbem.Uint32(job.priority)
    if job.started:
        model['StartTime'] = pywbem.CIMDateTime(datetime.fromtimestamp(
                job.started))
    model['TimeBeforeRemoval'] = pywbem.CIMDateTime(timedelta(
            seconds=job.time_before_removal))
    model['TimeOfLastStateChange'] = pywbem.CIMDateTime(datetime.fromtimestamp(
            job.last_change))
    model['TimeSubmitted'] = pywbem.CIMDateTime(datetime.fromtimestamp(
            job.created))

@cmpi_logging.trace_function
def job2model(job, class_name=None, keys_only=True, model=None):
    """
    Makes LMI_SoftwareJob out of job object or job id.

    :param job: (``int`` | ``YumAsyncJob``) Job identifier.
        In case of integer, caller should also provide class_name of resulting
        CIM instance. Otherwise generic LMI_SoftwareJob will be returned.
    :param class_name: (``str``) Determines CIM class name of resulting
        instance. This should be given when ``job`` is an integer.
    :param model: (``CIMInstance`` | ``CIMInstanceName``) If not None,
        will be filled with properties, otherwise
        a new instance of CIMInstance or CIMObjectPath is created.
    :param keys_only: (``bool``) Says whether to fill only key properties.
        Also if ``model`` is not given, it determines, whether to make
        ``CIMInstanceName`` or ``CIMInstance``.
    :rtype: (``CIMInstance`` | ``CIMInstanceName``)
    """
    if not isinstance(job, (int, long, jobs.YumAsyncJob)):
        raise TypeError("job must be an instance of YumAsyncJob")
    if isinstance(job, jobs.YumAsyncJob) and not job.async:
        raise ValueError("job must be asynchronous")
    if not keys_only and isinstance(job, (int, long)):
        raise TypeError("job must be an instance of YumAsyncJob"
            " filling non-key properties")

    if class_name is None:
        if  model is not None:
            class_name = model.classname
        elif isinstance(job, jobs.YumJob):
            class_name = job_class2cim_class_name(job.__class__)
        else:
            class_name = "LMI_SoftwareJob"
            LOG().warn("class_name not supplied for job %s, using general"
                    " LMI_SoftwareJob as CIM class name", job)
    if model is None:
        model = util.new_instance_name(class_name)
        if not keys_only:
            model = pywbem.CIMInstance(class_name, path=model)

    jobid = job.jobid if isinstance(job, jobs.YumAsyncJob) else job
    model['InstanceID'] = 'LMI:%s:%d' % (class_name, jobid)
    if isinstance(model, pywbem.CIMInstance):
        model.path['InstanceID'] = model['InstanceID']  #pylint: disable=E1103

    if not keys_only:
        _fill_nonkeys(job, model)
    return model

@cmpi_logging.trace_function
def object_path2job(op):
    """
    @param op must contain precise InstanceID of job
    """
    if not isinstance(op, pywbem.CIMInstanceName):
        raise pywbem.CIMError(pywbem.CIM_ERR_INVALID_PARAMETER,
                "op must be an instance of CIMInstanceName")

    if (not "InstanceID" in op or not op['InstanceID']):
        raise pywbem.CIMError(pywbem.CIM_ERR_INVALID_PARAMETER,
                "Missing InstanceID key property.")
    instid = op['InstanceID']
    match = util.RE_INSTANCE_ID.match(instid)
    if not match or match.group('clsname').lower() not in {
            c.lower() for c in JOB_CLASS_NAMES}:
        raise pywbem.CIMError(pywbem.CIM_ERR_INVALID_PARAMETER,
            "InstanceID must start with one of {%s} prefixes."
            " And end with positive integer." % (
                ", ".join(("LMI:%s:" % cn) for cn in JOB_CLASS_NAMES)))

    instid = int(match.group('id'))
    try:
        job = YumDB.get_instance().get_job(instid)
        clsname = job_class2cim_class_name(job.__class__)
        if (   clsname.lower() != op.classname.lower()
           and op.classname.lower() != 'LMI_SoftwareJob'.lower()):
            raise pywbem.CIMError(pywbem.CIM_ERR_INVALID_PARAMETER,
                "Classname \"%s\" does not belong to job with given id."
                " \"%s\" is the correct one." % (op.classname, clsname))
        return job
    except errors.JobNotFound:
        raise pywbem.CIMError(pywbem.CIM_ERR_NOT_FOUND,
            'No such job "%s".' % op['InstanceID'])

@cmpi_logging.trace_function
def modify_instance(instance):
    """
    This call modifies the job's parameters according to given instance.
    """
    job = object_path2job(instance.path)
    ydb = YumDB.get_instance()
    update_kwargs = {}
    reschedule_kwargs = {}
    # all modifiable properties
    prop_name_map = {
            "name"               : "name",
            "priority"           : "priority",
            "deleteoncompletion" : "delete_on_completion",
            "timebeforeremoval"  : "time_before_removal"
    }
    metadata_props = {"name"}
    reschedule_props = {"delete_on_completion", "time_before_removal"}
    for name, prop in instance.properties.items():
        if prop is None:
            LOG().warn('property "%s" is None', name)
            continue
        name = name.lower()
        try:
            pname = prop_name_map[name]
            if pname == "priority" and job.priority != prop.value:
                LOG().info('changing priority of job %s to %d', job, prop.value)
                job = ydb.set_job_priority(job.jobid, prop.value)
            elif pname in reschedule_props:
                if getattr(job, pname) == prop.value:
                    continue
                if pname == "time_before_removal":
                    value = prop.value.timedelta.total_seconds()
                else:
                    value = prop.value
                reschedule_kwargs[pname] = value
            else:
                if pname in metadata_props:
                    if not 'metadata' in update_kwargs:
                        update_kwargs['metadata'] = {}
                    update_kwargs['metadata'][pname] = prop.value
                else:
                    update_kwargs[pname] = prop.value
        except KeyError:
            if name == 'instanceid':
                continue
            LOG().warn("skipping property %s: %s", name, prop)

    if reschedule_kwargs:
        for prop in ('delete_on_completion', 'time_before_removal'):
            if prop not in reschedule_kwargs:
                reschedule_kwargs[prop] = getattr(job, prop)
        LOG().info('rescheduling job %s to: %s', job,
            ", ".join("%s=%s"%(k, v) for k, v in reschedule_kwargs.items()))
        job = ydb.reschedule_job(job.jobid, **reschedule_kwargs)

    if update_kwargs:
        LOG().info('changing atributes of job %s to: %s', job,
                ", ".join("%s=%s"%(k, v) for k, v in update_kwargs.items()))
        job = ydb.update_job(job.jobid, **update_kwargs)

    return job2model(job, keys_only=False, model=instance)

@cmpi_logging.trace_function
def job2error(env, job):
    """
    @return instance of CIM_Error if job is in EXCEPTION state,
    None otherwise
    """
    if not isinstance(job, jobs.YumJob):
        raise TypeError("job must be isntance of YumJob")
    if job.state == job.EXCEPTION:
        errortup = job.result_data
        kwargs = {}
        if issubclass(errortup[0],
                (errors.RepositoryNotFound, errors.PackageNotFound)):
            kwargs['status_code'] = Error.Values. \
                    CIMStatusCode.CIM_ERR_NOT_FOUND
            if issubclass(errortup[0], errors.PackageNotFound):
                kwargs['status_code_description'] = "Package not found"
            else:
                kwargs['status_code_description'] = "Repository not found"
        elif issubclass(errortup[0], errors.PackageAlreadyInstalled):
            kwargs['status_code'] = Error.Values. \
                    CIMStatusCode.CIM_ERR_ALREADY_EXISTS
        kwargs['message'] = getattr(errortup[1], 'message',
                str(errortup[1]))
        value = Error.make_instance(env, **kwargs)
        return value
